import threading

from pyserilog.capturing.iproperty_value_converter import IPropertyValueConverter
from pyserilog.core.ilog_event_property_value_factory import ILogEventPropertyValueFactory
from pyserilog.debuging.self_log import SelfLog
from pyserilog.events.log_event_property_value import LogEventPropertyValue
from pyserilog.events.scalar_value import ScalarValue
from pyserilog.parsing.destructuring import Destructuring


class DepthLimiter(ILogEventPropertyValueFactory):
    local = threading.local()

    def __init__(self, maximum_depth: int, property_value_converter: IPropertyValueConverter):
        self._maximum_depth = maximum_depth
        self._property_value_converter = property_value_converter

    @property
    def current_dept(self):
        return self.local.current_depth

    @staticmethod
    def set_current_depth(depth: int):
        loc = DepthLimiter.local
        loc.current_depth = depth

    def create_property_value(self, value, destructure_objects: bool = False) -> LogEventPropertyValue:
        stored_depth = self.current_dept

        created = self.__default_if_maximum_depth(stored_depth)
        if created is None:
            created = self._property_value_converter.create_property_value_check_destructuring_and_depth \
                (value, destructure_objects, stored_depth + 1)
        self.set_current_depth(stored_depth)
        return created

    # TODO refactor merge to object
    def create_property_value_with_destructure(self, value, destructuring: Destructuring) -> LogEventPropertyValue:
        stored_depth = self.current_dept

        created = self.__default_if_maximum_depth(stored_depth)
        if created is None:
            created = self._property_value_converter.create_property_value_with_destructuring_and_depth \
                (value, destructuring, stored_depth + 1)
        self.set_current_depth(stored_depth)
        return created

    def __default_if_maximum_depth(self, depth: int) -> ScalarValue:
        if depth == self._maximum_depth:
            SelfLog.write_line("Maximum destructuring depth reached.")
            return ScalarValue.Null()
        return None



from pyserilog.core.string_writable import StringWriteable

from pyserilog.parsing.alignment import Alignment, AlignmentDirection


class Padding:

    @staticmethod
    def apply(output: StringWriteable, value: str, text_alignment: Alignment | None):
        if text_alignment is None or len(value) >= text_alignment.width:
            output.write(value)
            return

        pad: int = text_alignment.width - len(value)
        if text_alignment.direction == AlignmentDirection.LEFT:
            output.write(value)
        output.write(' ' * pad)
        if text_alignment.direction == AlignmentDirection.RIGHT:
            output.write(value)



import sys

from pyserilog.capturing.message_template_processor import MessageTemplateProcessor
from pyserilog.capturing.property_value_converter import PropertyValueConverter
from pyserilog.configuration.logger_destructuring_configuration import LoggerDestructuringConfiguration
from pyserilog.configuration.logger_filter_configuration import LoggerFilterConfiguration
from pyserilog.configuration.logger_minimum_level_configuration import LoggerMinimumLevelConfiguration
from pyserilog.core.idestructuring_policy import IDestructuringPolicy
from pyserilog.core.ilog_event_enricher import ILogEventEnricher
from pyserilog.core.ilog_event_filter import ILogEventFilter
from pyserilog.core.ilog_event_sink import ILogEventSink
from pyserilog.core.level_override_map import LevelOverrideMap
from pyserilog.core.logger import Logger
from pyserilog.core.logging_level_switch import LoggingLevelSwitch
from pyserilog.core.enrichers.empty_enricher import EmptyEnricher
from pyserilog.core.enrichers.safe_aggregate_enricher import SafeAggregateEnricher
from pyserilog.core.sinks.aggregate_sink import AggregateSink
from pyserilog.core.sinks.filtering_sink import FilteringSink
from pyserilog.core.sinks.safe_aggregate_sink import SafeAggregateSink
from pyserilog.debuging import SelfLog
from pyserilog.events.log_event_level import LogEventLevel
from pyserilog.events.level_alias import LevelAlias
from pyserilog.configuration.logger_sink_configuration import LoggerSinkConfiguration
from pyserilog.configuration.logger_audit_sink_configuration import LoggerAuditSinkConfiguration
from pyserilog.configuration.logger_enrichment_configuration import LoggerEnrichmentConfiguration


class LoggerConfiguration:
    def __init__(self):
        self._log_event_sinks: list[ILogEventSink] = []
        self._audit_sinks: list[ILogEventSink] = []
        self._enrichers: list[ILogEventEnricher] = []
        self._filters: list[ILogEventFilter] = []
        self._additional_scalar_types: list[type] = []
        self._additional_destructuring_policies: list[IDestructuringPolicy] = []
        self._overrides: dict[str, LoggingLevelSwitch] = {}
        self._maximum_destructuring_depth = 10
        self._maximum_string_length = sys.maxsize
        self._maximum_collection_count = sys.maxsize

        self._write_to = LoggerSinkConfiguration(self, lambda s: self._log_event_sinks.append(s))
        self._enrich = LoggerEnrichmentConfiguration(self, lambda s: self._enrichers.append(s))

        self._logger_created = False
        self._minimum_level: LogEventLevel = LogEventLevel.INFORMATION
        self._level_switch: LoggingLevelSwitch | None = None

    @property
    def write_to(self) -> LoggerSinkConfiguration:
        """
        Configures the sinks that log events will be emitted to.
        """
        return self._write_to

    @property
    def audit_to(self) -> LoggerAuditSinkConfiguration:
        """
        Configures sinks for auditing, instead of regular (safe) logging. When auditing is used,
        exceptions from sinks and any intermediate filters propagate back to the caller. Most callers
        should use "write_to" instead

        Not all sinks are compatible with transactional auditing requirements (many will use asynchronous
        batching to improve write throughput and latency). Sinks need to opt-in to auditing support by
        extending "LoggerAuditSinkConfiguration", though the generic "LoggerAuditSinkConfiguration.Sink"
        method allows any sink class to be adapted for auditing.
        """

        def add_sink_function(event_sink: ILogEventSink):
            self._audit_sinks.append(event_sink)

        return LoggerAuditSinkConfiguration(self, add_sink_function=add_sink_function)

    @property
    def minimum_level(self) -> LoggerMinimumLevelConfiguration:
        def set_minimum_func(level: LogEventLevel):
            self._minimum_level = level
            self._level_switch = None

        def set_level_switch_func(level_switch: LoggingLevelSwitch):
            self._level_switch = level_switch

        def add_override_func(source: str, level_switch: LoggingLevelSwitch):
            self._overrides[source] = level_switch

        return LoggerMinimumLevelConfiguration(logger_configuration=self, set_minimum_func=set_minimum_func,
                                               set_level_switch_func=set_level_switch_func,
                                               add_override_func=add_override_func
                                               )

    @property
    def enrich(self) -> LoggerEnrichmentConfiguration:
        """
        Configures enrichment of <see cref="LogEvent"/>s. Enrichers can add, remove and
        modify the properties associated with events.
        :return:
        """
        return self._enrich

    @enrich.setter
    def enrich(self, value: LoggerEnrichmentConfiguration):
        self._enrich = value

    @property
    def filter(self) -> LoggerFilterConfiguration:
        def add_filter(log_event_filer: ILogEventFilter):
            self._filters.append(log_event_filer)

        return LoggerFilterConfiguration(self, add_filter)

    @property
    def destructure(self) -> LoggerDestructuringConfiguration:
        def set_maximum_depth_func(x: int):
            self._maximum_destructuring_depth = x

        def set_maximum_collection_count(x: int):
            self._maximum_collection_count = x

        def set_maximum_string_length(x: int):
            self._maximum_string_length = x

        def add_scalar_func(scalar_type: type):
            self._additional_scalar_types.append(scalar_type)

        return LoggerDestructuringConfiguration(
            logger_configuration=self,
            add_scalar_func=add_scalar_func,
            add_policy_func=self._additional_destructuring_policies.append,
            set_maximum_depth_func=set_maximum_depth_func,
            set_maximum_collection_count=set_maximum_collection_count,
            set_maximum_string_length=set_maximum_string_length
        )

    @property
    def read_from(self):
        from pyserilog.configuration.logger_settings_configuration import LoggerSettingsConfiguration
        return LoggerSettingsConfiguration(self)


    def create_logger(self) -> Logger:
        if self._logger_created:
            raise PermissionError("create_logger() was previously called and can only be called once.")

        self._logger_created = True

        sink = self.__get_sink()
        enricher = self.__get_enricher()
        overrider_map = self.__get_overrider_map()
        processor = self.__get_template_processor()
        exit_func = self.__exit_sinks_func

        minimum_level = LevelAlias.minimum if self._level_switch is not None else self._minimum_level
        return Logger(processor, minimum_level, self._level_switch, sink, enricher, exit_func, overrider_map)

    def __get_template_processor(self):
        propagate_exception = len(self._audit_sinks) > 0
        converter = PropertyValueConverter(self._maximum_destructuring_depth,
                                           self._maximum_string_length,
                                           self._maximum_collection_count,
                                           self._additional_scalar_types,
                                           self._additional_destructuring_policies,
                                           propagate_exception
                                           )
        processor = MessageTemplateProcessor(converter)
        return processor

    def __get_enricher(self):
        match len(self._enrichers):
            case 0:
                return EmptyEnricher()
            case 1:
                return self._enrichers[0]
            case _:
                return SafeAggregateEnricher(self._enrichers)

    def __get_overrider_map(self) -> LevelOverrideMap:
        if len(self._overrides) > 0:
            return LevelOverrideMap(self._overrides, self._minimum_level, self._level_switch)
        return None

    def __exit_sinks_func(self, exc_type, exc_val, exc_tb):
        all_handlers = []
        all_handlers.extend(self._log_event_sinks)
        all_handlers.extend(self._audit_sinks)

        for hand in all_handlers:
            try:
                if hasattr(hand, "__exit__"):
                    hand.__exit__(exc_type, exc_val, exc_tb)
            except Exception as ex:
                SelfLog.write_line("error {0} happened in exit {1} ", ex, hand)

    def __get_sink(self) -> ILogEventSink:
        sink: ILogEventSink = None
        if len(self._log_event_sinks) > 0:
            sink = SafeAggregateSink(self._log_event_sinks)

        auditing = len(self._audit_sinks) > 0
        if auditing:
            sinks = []
            if sink is not None:
                sinks.append(sink)
            sinks.extend(self._audit_sinks)
            sink = AggregateSink(sinks)

        if sink is None:
            sink = SafeAggregateSink([])

        if len(self._filters) > 0:
            sink = FilteringSink(sink, self._filters, auditing)

        return sink



from enum import Enum


class Destructuring(Enum):
    """
    Instructs the logger on how to store information about provided parameters.
    """
    DEFAULT = 0
    """
    Convert known types and objects to scalars, arrays to sequences.
    """
    STRINGIFY = 1
    """
    Convert all types to scalar strings. Prefix name with '$'.
    """
    DESTRUCTURE = 2
    """
    Convert known types to scalars, destructure objects and collections 
    into sequences and structures. Prefix name with '@'.
    """


from abc import ABC, abstractmethod

from pyserilog.core.ilog_event_property_value_factory import ILogEventPropertyValueFactory
from pyserilog.events.log_event_property_value import LogEventPropertyValue


class IDestructuringPolicy(ABC):
    @abstractmethod
    def try_destructure(self, value, property_value_factory: ILogEventPropertyValueFactory) \
            -> tuple[bool, LogEventPropertyValue | None]:
        pass



from pyserilog.guard import Guard


def is_valid_name(name: str):
    return name is not None and len(name) > 0


def ensure_valid_name(name: str):
    Guard.against_null(name)
    if not is_valid_name(name):
        raise Exception("Property name must not be empty or whitespace.")



from enum import Enum;


class LogEventLevel(Enum):
    """
    Specifies the meaning and relative importance of a log event.
    """
    VERBOSE = 0,
    """
    Anything and everything you might want to know about a running block of code.
    """
    DEBUG = 1,
    """
    Internal system events that aren't necessarily observable from the outside.
    """
    INFORMATION = 2,
    """
    The lifeblood of operational intelligence - things happen.
    """
    WARNING = 3,
    """
    Service is degraded or endangered.
    """
    ERROR = 4,
    """
    Functionality is unavailable, invariants are broken or data is lost.
    """
    FATAL = 5,
    """
    If you have a pager, it goes off when one of these occurs.

    """

    def get_int(self):
        return self.value[0]

    def __gt__(self, other):
        if isinstance(other, LogEventLevel):
            return self.value > other.value

        raise Exception("Invalid Compare")

    def __ge__(self, other):
        if isinstance(other, LogEventLevel):
            return self.value >= other.value
        raise Exception("Invalid Compare")

    def __add__(self, other):
        if isinstance(other, int):
            val = other + self.get_int()
            t = self.__convert_from_int(val)
            return t
        if isinstance(other, LogEventLevel):
            val1 = other.get_int()
            val2 = self.get_int()
            return self.__convert_from_int(val1 + val2)
        raise TypeError()

    @staticmethod
    def __convert_from_int(val: int):
        for p in LogEventLevel:
            if p.value[0] == val:
                return p
        return ValueError("invalid range number")



from pyserilog.core.ilog_event_enricher import ILogEventEnricher
from pyserilog.core.ilog_event_property_factory import ILogEventPropertyFactory
from pyserilog.debuging.self_log import SelfLog
from pyserilog.events.log_event import LogEvent


class SafeAggregateEnricher(ILogEventEnricher):
    def __init__(self, enrichers: list[ILogEventEnricher]):
        self._enrichers = enrichers

    def enrich(self, log_event: LogEvent, property_factory: ILogEventPropertyFactory):
        for enricher in self._enrichers:
            try:
                enricher.enrich(log_event, property_factory)
            except Exception as ex:
                SelfLog.write_line("Exception {0} caught while enriching {1} with {2}.", ex, log_event, enricher)



from abc import ABC, abstractclassmethod, abstractmethod

from pyserilog import LoggerConfiguration


class ILoggerSettings(ABC):

    @abstractmethod
    def configure(self, logger_configuration: LoggerConfiguration):
        pass



from abc import ABC, abstractmethod

from pyserilog.events import message_template


class IMessageTemplateParser(ABC):

    @abstractmethod
    def parse(self, message_template: str) -> message_template:
        pass



from pyserilog.core.string_writable import StringWriteable
from typing import Callable

from pyserilog.events.log_event_property_value import LogEventPropertyValue


class ScalarValue(LogEventPropertyValue):
    def __init__(self, value):
        self._value = value

    @staticmethod
    def Null():
        return ScalarValue(None)

    @property
    def value(self):
        return self._value

    def render(self, output: StringWriteable, text_format: str | None = None,
               formatter: Callable[[object, str | None], str] = None):
        self.render_value(self.value, output, text_format, formatter)

    @staticmethod
    def render_value(value, writer: StringWriteable, text_format: str | None = None,
                     formatter: Callable[[object, str | None], str] = None):
        if value is None:
            writer.write("null")
        elif isinstance(value, str):
            writer.write(f'"{value}"')
        else:
            if text_format is not None:
                try:
                    format_res = format(value, text_format) if formatter is None else formatter(value, text_format)
                    writer.write(format_res)
                except Exception as ex:
                    print(ex)
            else:
                writer.write(f"{value}")

    @staticmethod
    def Null():
        return ScalarValue(None)

    def __eq__(self, other):
        if not isinstance(other, ScalarValue):
            return False
        return other._value == self.value



from typing import Callable

from pyserilog.core.ilog_event_filter import ILogEventFilter
from pyserilog.events.log_event import LogEvent


class DelegateFilter(ILogEventFilter):
    def __init__(self, is_enable_func: Callable[[LogEvent], bool]):
        self._is_enable_func = is_enable_func

    def is_enable(self, log_event: LogEvent) -> bool:
        return self._is_enable_func(log_event)



from builtins import ExceptionGroup

from pyserilog import Guard
from pyserilog.core.ilog_event_sink import ILogEventSink
from pyserilog.debuging import SelfLog
from pyserilog.events.log_event import LogEvent


class DisposingAggregateSink(ILogEventSink):
    def __init__(self, sinks: list[ILogEventSink]):
        Guard.against_null(sinks)
        self._sinks = sinks

    def emit(self, log_event: LogEvent):
        exceptions = []
        for sink in self._sinks:
            try:
                sink.emit(log_event)
            except Exception as ex:
                SelfLog.write_line("Caught exception while emitting to sink {0}: {1}", sink, ex)
                exceptions.append(ex)

        if len(exceptions) > 0:
            raise ExceptionGroup(exceptions)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for sink in self._sinks:
            try:
                if hasattr(sink, "__exit__"):
                    sink.__exit__(exc_type, exc_val, exc_tb)
            except Exception as ex:
                SelfLog.write_line("Caught exception while disposing sink {0}: {1}", sink, ex)



import inspect
from inspect import FullArgSpec
from types import ModuleType

import pyserilog.settings.key_value_paies.surrogate_configuration_methods as sm


class MethodInfo:
    def __init__(self, klass_type: type, method_name: str, func, args: FullArgSpec):
        self.klass_type = klass_type
        self.method_name = method_name
        self.args = args
        self.func = func

    def __str__(self):
        return f"{self.klass_type} , {self.method_name} , {self.func}"


class CallableConfigurationMethodFinder:

    @staticmethod
    def find_configuration_methods(modules: list[ModuleType], config_type: type):
        all_modules = [sm]
        all_modules.extend(modules)
        result = []
        for module in all_modules:
            all_module_klass = inspect.getmembers(module, predicate=inspect.isclass)
            for _, klass_type in all_module_klass:
                methods = inspect.getmembers_static(klass_type, predicate=lambda x: isinstance(x, staticmethod))
                for method_name, _ in methods:
                    func = getattr(klass_type, method_name)
                    args = inspect.getfullargspec(func)

                    if len(args.args) == 0:
                        continue
                    first_arg_name = args.args[0]
                    if len(args.annotations) > 0 and first_arg_name in args.annotations and \
                            args.annotations[first_arg_name] == config_type:
                        info = MethodInfo(klass_type, method_name, func, args)
                        result.append(info)
        return result



from pyserilog import Guard
from pyserilog.core import logging_level_switch
from pyserilog.events import log_event_level
from pyserilog.events.level_alias import LevelAlias
from pyserilog.events.log_event_level import LogEventLevel


class LevelOverrideMap:
    class LevelOverride:
        def __init__(self, context: str, level_switch: logging_level_switch):
            self._context = context
            self._level_switch = level_switch

        @property
        def context(self):
            return self._context

        @property
        def level_switch(self):
            return self._level_switch

        def __str__(self):
            return f"{{context :{self.__class__} , overrider: {self._level_switch}}}"

    def __init__(self, overrides: dict[str, logging_level_switch], default_minimum_level: LogEventLevel,
                 default_level_switch: logging_level_switch):
        Guard.against_null(overrides)

        self._default_level_switch = default_level_switch
        self._default_minimum_level = LevelAlias.minimum if default_level_switch is not None else default_minimum_level

        keys = list(overrides.keys())
        keys.sort(reverse=True)

        self._overrides: list[LevelOverrideMap.LevelOverride] = []
        for key in keys:
            self._overrides.append(LevelOverrideMap.LevelOverride(key, overrides[key]))

    def get_effective_level(self, context: str) -> tuple[log_event_level, logging_level_switch]:
        for level_override in self._overrides:
            if context.startswith(level_override.context) and (
                    len(context) == len(level_override.context) or context[len(level_override.context)] == '.'):
                return LevelAlias.minimum, level_override.level_switch

        return self._default_minimum_level, self._default_level_switch





import importlib
from ctypes import FormatError
from types import ModuleType
from typing import Callable

from pyserilog import LoggerConfiguration, Guard
from pyserilog.configuration.ilogger_settings import ILoggerSettings
from pyserilog.configuration.logger_audit_sink_configuration import LoggerAuditSinkConfiguration
from pyserilog.configuration.logger_destructuring_configuration import LoggerDestructuringConfiguration
from pyserilog.configuration.logger_enrichment_configuration import LoggerEnrichmentConfiguration
from pyserilog.configuration.logger_filter_configuration import LoggerFilterConfiguration
from pyserilog.configuration.logger_sink_configuration import LoggerSinkConfiguration
from pyserilog.core.logging_level_switch import LoggingLevelSwitch
from pyserilog.debuging import SelfLog
from pyserilog.events.log_event_level import LogEventLevel
import re

from pyserilog.settings.key_value_paies.callable_configuration_method_finder import CallableConfigurationMethodFinder, \
    MethodInfo
from pyserilog.settings.key_value_paies.setting_value_conversions import SettingValueConversions


class KeyValuePairSettings(ILoggerSettings):
    USING_DIRECTIVE = "using"
    LEVEL_SWITCH_DIRECTIVE = "level-switch"
    AUDIT_TO_DIRECTIVE = "audit-to"
    WRITE_TO_DIRECTIVE = "write-to"
    MINIMUM_LEVEL_DIRECTIVE = "minimum-level"
    MINIMUM_LEVEL_CONTROLLED_BY_DIRECTIVE = "minimum-level:controlled-by"
    ENRICH_WITH_DIRECTIVE = "enrich"
    ENRICH_WITH_PROPERTY_DIRECTIVE = "enrich:with-property"
    FILTER_DIRECTIVE = "filter"
    DESTRUCTURE_DIRECTIVE = "destructure"

    USING_DIRECTIVE_FULL_FORM_PREFIX = "using:"
    ENRICH_WITH_PROPERTY_DIRECTIVE_PREFIX = "enrich:with-property:"
    MINIMUM_LEVEL_OVERRIDE_DIRECTIVE_PREFIX = "minimum-level:override:"

    CALLABLE_DIRECTIVE_REGEX = '^(?P<directive>audit-to|write-to|enrich|filter|destructure):(?P<method>[A-Za-z0-9_]*)' \
                               '(\\.(?P<argument>[A-Za-z0-9_]*)){0,1}'
    LEVEL_SWITCH_DECLARATION_DIRECTIVE_REGEX = "^level-switch:(?P<switch_name>.*)"
    LEVEL_SWITCH_NAME_REGEX = '^\\$[A-Za-z]+[A-Za-z0-9]*$'

    _supportedDirectives = [
        USING_DIRECTIVE,
        LEVEL_SWITCH_DIRECTIVE,
        AUDIT_TO_DIRECTIVE,
        WRITE_TO_DIRECTIVE,
        MINIMUM_LEVEL_DIRECTIVE,
        MINIMUM_LEVEL_CONTROLLED_BY_DIRECTIVE,
        ENRICH_WITH_PROPERTY_DIRECTIVE,
        ENRICH_WITH_DIRECTIVE,
        FILTER_DIRECTIVE,
        DESTRUCTURE_DIRECTIVE
    ]

    callable_directive_receiver_types: dict[str, type] = {
        "audit-to": LoggerAuditSinkConfiguration,
        "write-to": LoggerSinkConfiguration,
        "enrich": LoggerEnrichmentConfiguration,
        "filter": LoggerFilterConfiguration,
        "destructure": LoggerDestructuringConfiguration
    }
    callable_directive_receivers: dict[type, Callable[[LoggerConfiguration], object]] = {
        LoggerAuditSinkConfiguration: lambda lc: lc.audit_to,
        LoggerSinkConfiguration: lambda lc: lc.write_to,
        LoggerEnrichmentConfiguration: lambda lc: lc.enrich,
        LoggerFilterConfiguration: lambda lc: lc.filter,
        LoggerDestructuringConfiguration: lambda lc: lc.destructure,
    }

    def __init__(self, settings: dict[str, str]):
        self._settings = settings

    def configure(self, logger_configuration: LoggerConfiguration):
        Guard.against_null(logger_configuration)

        directives = {k: v for k, v in self._settings.items() if
                      any(filter(lambda x: k.startswith(x), self._supportedDirectives))}

        declared_level_switches = self.__parse_named_level_switch_declaration_directives(directives)

        if self.MINIMUM_LEVEL_DIRECTIVE in directives:
            minimum_level_directive = directives[self.MINIMUM_LEVEL_DIRECTIVE]
            if minimum_level_directive in LogEventLevel._member_names_:
                min_level = LogEventLevel[minimum_level_directive]
                logger_configuration.minimum_level.level_is(min_level)

        self.__config_minimum_level_controlled_by(directives, logger_configuration, declared_level_switches)
        self.__config_enrich_with_property(directives, logger_configuration)
        self.__config_minimum_level_with_override_prefix(directives, logger_configuration, declared_level_switches)
        self.__config_callable_methods(directives, logger_configuration, declared_level_switches)
        pass

    def __config_enrich_with_property(self, directives: dict[str, str], logger_configuration: LoggerConfiguration):
        def check_enrich_property_with_name(key_enrich: str):
            start_with = key_enrich.startswith(self.ENRICH_WITH_PROPERTY_DIRECTIVE_PREFIX)
            return start_with and len(key_enrich) > len(self.ENRICH_WITH_PROPERTY_DIRECTIVE_PREFIX)

        for key in filter(check_enrich_property_with_name, directives):
            name = key[len(self.ENRICH_WITH_PROPERTY_DIRECTIVE_PREFIX):]
            value = directives[key]
            logger_configuration.enrich.with_property(name, value)

    def __config_minimum_level_with_override_prefix(self, directives: dict[str, str],
                                                    logger_configuration: LoggerConfiguration, declared_level_switches):
        def is_contain(level_key: str):
            start_with = level_key.startswith(self.MINIMUM_LEVEL_OVERRIDE_DIRECTIVE_PREFIX)
            return start_with and len(level_key) > len(self.MINIMUM_LEVEL_OVERRIDE_DIRECTIVE_PREFIX)

        for key in filter(is_contain, directives):
            package_prefix = key[len(self.MINIMUM_LEVEL_OVERRIDE_DIRECTIVE_PREFIX):]
            value = directives[key]
            if value in LogEventLevel._member_names_:
                level = LogEventLevel[value]
            else:
                level = self.__look_up_switch_by_name(value, declared_level_switches)
            logger_configuration.minimum_level.override(package_prefix, level)

    def __config_callable_methods(self, directives: dict[str, str], logger_configuration: LoggerConfiguration,
                                  declared_level_switches):
        res = []

        for key in directives:
            match = re.match(self.CALLABLE_DIRECTIVE_REGEX, key)
            if match:
                received_type = self.callable_directive_receiver_types[match.group("directive")]
                method = match.group("method")
                argument = match.group("argument")
                call = KeyValuePairSettings.ConfigurationMethodCall(method, argument, directives[key])
                call_res = KeyValuePairSettings.CallableResult(received_type, call)
                res.append(call_res)
        if not any(res):
            return

        modules = self.__load_configuration_modules(directives)
        receiver_types = set(list(map(lambda x: x.receiver_type, res)))
        for receiver_type in receiver_types:
            methods = CallableConfigurationMethodFinder.find_configuration_methods(modules, receiver_type)
            method_names = set(
                map(lambda x: x.call.method_name, filter(lambda x: x.receiver_type == receiver_type, res)))
            for method_name in method_names:
                target: MethodInfo = self.__select_configuration_method(methods, method_name)
                if target is None:
                    SelfLog.write_line('Setting "{0}" could not be matched to an implementation in any of the loaded '
                                       'assemblies. To use settings from additional assemblies, specify them with the'
                                       ' "serilog:import" key.')
                else:
                    kwargs = KeyValuePairSettings.__get_method_args(method_name, receiver_type, res, target,
                                                                    declared_level_switches)
                    obj = self.callable_directive_receivers[receiver_type](logger_configuration)
                    kwargs[target.args.args[0]] = obj
                    target.func(**kwargs)

    @staticmethod
    def __get_method_args(method_name: str, receiver_type: type, callable_results, target: MethodInfo,
                          declared_level_switches):
        callable_results: list[KeyValuePairSettings.CallableResult] = callable_results
        res: list[KeyValuePairSettings.CallableResult] = list(
            filter(lambda x: x.call.method_name == method_name and x.receiver_type == receiver_type, callable_results)
        )

        kwargs = {}
        for receiver in res:
            arg_name = receiver.call.argument_name
            arg_type = target.args.annotations[arg_name]
            arg_value = KeyValuePairSettings.__convert_or_lookup_by_name(receiver.call.value, arg_type,
                                                                         declared_level_switches)
            kwargs[arg_name] = arg_value
        return kwargs

    @staticmethod
    def __convert_or_lookup_by_name(value_or_switch_name: str, klass_type: type,
                                    declared_switches: dict[str, LoggingLevelSwitch]):
        if issubclass(klass_type, LoggingLevelSwitch):
            return KeyValuePairSettings.__look_up_switch_by_name(value_or_switch_name, declared_switches)
        return SettingValueConversions.convert_to_type(value_or_switch_name, klass_type)

    def __load_configuration_modules(self, directives: dict[str, str]) -> list[ModuleType]:
        modules = set()

        def func(x: str):
            return x == self.USING_DIRECTIVE or x.startswith(self.USING_DIRECTIVE_FULL_FORM_PREFIX)

        for import_directive in filter(func, directives):
            value: str = directives[import_directive]
            if len(value.strip()) == 0:
                raise ValueError(
                    "A zero-length or whitespace module name was supplied to a serilog:import configuration statement.")
            module = importlib.import_module(value)
            modules.add(module)
        return list(modules)

    def __config_minimum_level_controlled_by(self, directives: dict[str, str],
                                             logger_configuration: LoggerConfiguration,
                                             declared_level_switches: dict[str, LoggingLevelSwitch]):
        if self.MINIMUM_LEVEL_CONTROLLED_BY_DIRECTIVE in directives:
            switch_name = directives[self.MINIMUM_LEVEL_CONTROLLED_BY_DIRECTIVE]
            global_minimum_level_switch = self.__look_up_switch_by_name(switch_name, declared_level_switches)
            logger_configuration.minimum_level.controlled_by(global_minimum_level_switch)

    @staticmethod
    def __select_configuration_method(candidate_methods: list[MethodInfo], name: str):

        def is_valid(x: MethodInfo):
            return x.method_name == name

        result = list(filter(lambda x: x.method_name == name, candidate_methods))
        if len(result) > 0:
            return result[0]

    @staticmethod
    def __parse_named_level_switch_declaration_directives(directives: dict[str, str]):
        res = re.compile(KeyValuePairSettings.LEVEL_SWITCH_DECLARATION_DIRECTIVE_REGEX)

        switches = []
        for key in directives:
            match = res.match(key)
            if match:
                switch_name = match.group("switch_name")
                initial_switch_level = directives[key]
                switches.append((switch_name, initial_switch_level))
        named_switches: dict[str, LoggingLevelSwitch] = {}
        for switch_name, initial_switch_level in switches:
            if not KeyValuePairSettings.is_valid_switch_name(switch_name):
                raise ValueError(f'"{switch_name}" is not a valid name for a Level Switch declaration. '
                                 f'Level switch must be declared with a \'$\' sign, like "level-switch:$switchName"')

            if initial_switch_level == "":
                named_switches[switch_name] = LoggingLevelSwitch()
            else:
                initial_level = LogEventLevel[initial_switch_level]
                named_switches[switch_name] = LoggingLevelSwitch(initial_level)
        return named_switches

    @staticmethod
    def __look_up_switch_by_name(switch_name: str,
                                 declared_level_switches: dict[str, LoggingLevelSwitch]) -> LoggingLevelSwitch:
        if switch_name in declared_level_switches:
            return declared_level_switches[switch_name]
        raise ValueError(f'No LoggingLevelSwitch has been declared with name "{switch_name}\".'
                         f' You might be missing a key "{KeyValuePairSettings.LEVEL_SWITCH_DIRECTIVE}:{switch_name}"')

    @staticmethod
    def is_valid_switch_name(inp: str):
        res = re.search(KeyValuePairSettings.LEVEL_SWITCH_NAME_REGEX, inp)
        return True if res else False

    class CallableResult(object):
        def __init__(self, receiver_type: type, call):
            self.receiver_type = receiver_type
            self.call: KeyValuePairSettings.ConfigurationMethodCall = call

    class ConfigurationMethodCall:
        def __init__(self, method_name: str, argument_name: str, value: str):
            self.method_name = method_name
            self.argument_name = argument_name
            self.value = value



from pyserilog.guard import Guard
from pyserilog.core.ilog_event_enricher import ILogEventEnricher
from pyserilog.core.ilog_event_property_factory import ILogEventPropertyFactory
from pyserilog.events.log_event import LogEvent
from pyserilog.events.event_property import EventProperty


class FixedPropertyEnricher(ILogEventEnricher):
    def __init__(self, event_property: EventProperty):
        self._event_property = event_property

    def enrich(self, log_event: LogEvent, property_factory: ILogEventPropertyFactory):
        Guard.against_null(log_event)

        log_event.add_property_if_absent(self._event_property)



from builtins import ExceptionGroup

from pyserilog.guard import Guard
from pyserilog.core.ilog_event_sink import ILogEventSink
from pyserilog.debuging.self_log import SelfLog
from pyserilog.events.log_event import LogEvent


class AggregateSink(ILogEventSink):
    def __init__(self, sinks: list[ILogEventSink]):
        Guard.against_null(sinks)
        self._sinks = sinks

    def emit(self, log_event: LogEvent):
        exceptions: list[Exception] = []

        for sink in self._sinks:
            try:
                sink.emit(log_event)
            except Exception as ex:
                SelfLog.write_line("Caught exception while emitting to sink {0}: {1}", sink, ex)
                exceptions.append(ex)

        if len(exceptions) > 0:
            raise ExceptionGroup("Failed to emit a log event.", exceptions)



import decimal
import math
from datetime import timedelta, datetime, date

from pyserilog.core.string_writable import StringIOWriteable, StringWriteable
from pyserilog.events.log_event_property import LogEventProperty

from pyserilog.guard import Guard
from pyserilog.data.log_event_property_value_visitor import LogEventPropertyValueVisitor
from pyserilog.events import log_event_property
from pyserilog.events.dictionary_value import DictionaryValue
from pyserilog.events.log_event_property_value import LogEventPropertyValue
from pyserilog.events.scalar_value import ScalarValue
from pyserilog.events.sequence_value import SequenceValue
from pyserilog.events.structure_value import StructureValue


class JsonValueFormatter(LogEventPropertyValueVisitor[StringWriteable, bool]):
    def __init__(self, type_tag_name: str | None = "_typeTag"):
        self._type_tag_name = type_tag_name

    def format(self, value: LogEventPropertyValue, output: StringWriteable):
        self.visit(output, value)

    def _visit_scalar_value(self, state: StringWriteable, value: ScalarValue) -> bool:
        Guard.against_null(value)
        self._format_literal_value(value.value, state)
        return False

    def _visit_sequence_value(self, output: StringWriteable, sequence: SequenceValue) -> bool:
        Guard.against_null(sequence)

        output.write("[")
        for element in sequence.elements[:-1]:
            self.visit(output, element)
            output.write(",")
        if len(sequence.elements) > 0:
            self.visit(output, sequence.elements[-1])
        output.write("]")
        return False

    def _visit_structure_value(self, state: StringWriteable, structure: StructureValue) -> bool:
        Guard.against_null(structure)

        def write_property(event_property: LogEventProperty):
            self.write_quoted_json_string(value=event_property.name, output=state)
            state.write(":")
            self.visit(state=state, value=event_property.value)

        state.write("{")
        if len(structure.properties) > 0:
            write_property(structure.properties[0])

        for prop in structure.properties[1:]:
            state.write(",")
            write_property(prop)

        if self._type_tag_name is not None and structure.type_tag is not None:
            if len(structure.properties) > 0:
                state.write(",")
            self.write_quoted_json_string(value=self._type_tag_name, output=state)
            state.write(":")
            self.write_quoted_json_string(value=structure.type_tag, output=state)

        state.write("}")
        return False

    def _visit_dictionary_value(self, state: StringWriteable, dictionary: DictionaryValue) -> bool:
        Guard.against_null(dictionary)

        state.write("{")
        delim = None
        for (element_key, element_value) in dictionary.elements:

            if delim is not None:
                state.write(delim)
            delim = ","
            val = str(element_key) if element_key is not None else "null"
            self.write_quoted_json_string(value=val, output=state)
            state.write(":")
            self.visit(state=state, value=element_value)
        state.write("}")
        return False

    def _format_literal_value(self, value, output: StringWriteable):
        if value is None:
            self._format_null_value(output)
            return

        if isinstance(value, str):
            self._format_string_value(output, value)
            return

        if isinstance(value, bool):
            data = 'true' if value else 'false'
            output.write(data)
            return

        if isinstance(value, int) or isinstance(value, float):
            self._format_number_value(value=value, output=output)
            return

        if isinstance(value, timedelta):
            data = self.td_format(value)
            self._format_string_value(output=output, value=data)
            return
        if isinstance(value, datetime):
            res = self.datetime_format(value)
            self._format_string_value(output=output, value=res)
            return

        if isinstance(value, decimal.Decimal):
            res = "{0:f}".format(value)
            output.write(res)
            return
        self._format_literal_object_value(value, output)

    @staticmethod
    def _format_null_value(output: StringWriteable):
        output.write("null")

    @staticmethod
    def _format_string_value(output: StringWriteable, value: str):
        JsonValueFormatter.write_quoted_json_string(value, output)

    @staticmethod
    def write_quoted_json_string(value: str, output: StringWriteable):
        output.write('"')

        clean_segment_start = 0
        any_escaped = False

        for i in range(len(value)):
            ch = value[i]
            if ord(ch) < 32 or ch == '\\' or ch == '"':
                any_escaped = True

                output.write(value[clean_segment_start: i])

                clean_segment_start = i + 1

                match ch:
                    case '"':
                        output.write('\\\"')
                    case '\\':
                        output.write('\\\\')
                    case '\n':
                        output.write('\\n')
                    case '\r':
                        output.write('\\r')
                    case '\f':
                        output.write('\\f')
                    case '\t':
                        output.write('\\t')
                    case _:
                        output.write('\\u')
                        output.write(format(ord(ch), "04X"))

        if any_escaped:
            output.write(value[clean_segment_start:])
        else:
            output.write(value)
        output.write('\"')

    def _format_literal_object_value(self, value, output: StringWriteable):
        Guard.against_null(value)

        res = str(value)
        string_value = "" if res is None else res
        self._format_string_value(output, string_value)

    def _format_number_value(self, value, output: StringWriteable):
        if math.isnan(value):
            self._format_string_value(output=output, value="NaN")
        elif math.isinf(value):
            if value > 0:
                self._format_string_value(output=output, value="Infinity")
            else:
                self._format_string_value(output=output, value="-Infinity")
        else:
            output.write(str(value))
        return

    @staticmethod
    def datetime_format(datetime_object: datetime):
        res = datetime_object.strftime("%Y-%m-%dT%H:%M:%S.%f%z")
        if datetime_object.tzinfo is not None:
            res = f"{res[:-2]}:{res[-2:]}"
        return res

    @staticmethod
    def td_format(td_object: timedelta):
        is_negative = False
        seconds = int(td_object.total_seconds())
        if seconds < 0:
            is_negative = True
            seconds = seconds * -1
        periods = [
            ('day', 60 * 60 * 24),
            ('hour', 60 * 60),
            ('minute', 60),
            ('second', 1)
        ]

        writer = StringIOWriteable()
        if is_negative:
            writer.write("-")
        for i in range(4):
            (period_name, period_seconds) = periods[i]
            if i == 0 and period_seconds > seconds:
                continue
            period_value, seconds = divmod(seconds, period_seconds)
            if i == 0:
                writer.write(str(period_value))
            else:
                writer.write(format(period_value, "02d"))

            if i == 0:
                writer.write(".")
            elif i != 3:
                writer.write(":")

        res = writer.getvalue()
        writer.close()
        return res



import enum
import typing
from typing import Callable, get_args
from datetime import date, time, datetime
import pydoc as pydoc
import inspect

from pyserilog.debuging import SelfLog


class SettingValueConversions:
    __converters: dict[type, Callable[[str], object]] = {
        int: lambda x: None if x is None or x == "" else int(x),
        float: lambda x: None if x is None or x == "" else float(x),
        str: lambda x: x,
        time: lambda x: time.fromisoformat(x),
        date: lambda x: date.fromisoformat(x),
        datetime: lambda x: datetime.fromisoformat(x),
        type: lambda x: pydoc.locate(x)
    }

    @staticmethod
    def convert_to_type(value: str, klass_type: type):
        if klass_type in SettingValueConversions.__converters:
            converter = SettingValueConversions.__converters[klass_type]
            return converter(value)

        if issubclass(klass_type, enum.Enum):
            return klass_type.__members__[value]
        elif issubclass(klass_type, list) or typing.get_origin(klass_type) is list:
            items = value.split(',')
            res = []
            if len(value) == 0:
                return res
            element_type = get_args(klass_type)[0]
            for item in items:
                converted_item = SettingValueConversions.convert_to_type(item, element_type)
                res.append(converted_item)
            return res

        if value is not None and value != "":
            t: type = pydoc.locate(value)
            if t is not None:
                ins = t()
                return ins
        return value



from pyserilog.core.string_writable import StringWriteable

from pyserilog.events.log_event_property import LogEventProperty
from pyserilog.events.log_event_property_value import LogEventPropertyValue
from pyserilog.events.message_template import MessageTemplate
from pyserilog.events.structure_value import StructureValue
from pyserilog.formatting.json.json_value_formatter import JsonValueFormatter


class PropertiesOutputFormat:
    formatter: JsonValueFormatter = JsonValueFormatter("$type")

    @staticmethod
    def render(template: MessageTemplate, properties: dict[str, LogEventPropertyValue],
               output_template: MessageTemplate, output: StringWriteable, text_format: str | None = None):

        def contain_property(prop_key):
            return PropertiesOutputFormat.__template_contain_property_name(template, prop_key) or \
                   PropertiesOutputFormat.__template_contain_property_name(output_template, prop_key)

        if text_format is not None and "j" in text_format:
            props = list(map(lambda x: LogEventProperty(x, properties[x]),
                             filter(lambda x: not contain_property(x), properties)))
            sv = StructureValue(props)
            PropertiesOutputFormat.formatter.format(sv, output)
            return

        output.write("{ ")
        delim = ""
        for key in properties:
            if contain_property(key):
                continue
            output.write(delim)
            delim = ", "
            output.write(key)
            output.write(": ")
            properties[key].render(output, None)
        output.write(" }")

    @staticmethod
    def __template_contain_property_name(template: MessageTemplate, property_name: str):
        if template.positional_properties is not None:
            for p in template.positional_properties:
                if p.property_name == property_name:
                    return True
            return False

        if template.named_properties is not None:
            for p in template.named_properties:
                if p.property_name == property_name:
                    return True
            return False
        return False



from typing import Callable

from pyserilog.core.ilog_event_sink import ILogEventSink
from pyserilog.events.log_event import LogEvent


class DisposeDelegatingSink(ILogEventSink):
    def __init__(self, sink: ILogEventSink, exitable_func):
        self._sink = sink
        self._exitable = exitable_func

    def emit(self, log_event: LogEvent):
        self._sink.emit(log_event)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._exitable is not None and hasattr(self._exitable , "__exit__"):
            self._exitable.__exit__(exc_type, exc_val, exc_tb)



from pyserilog.events.level_alias import LevelAlias
from pyserilog.events.log_event_level import LogEventLevel
from pyserilog.rendering.casing import Casing


class LevelOutputFormat:
    _titleCaseLevelMap = [
        ["V", "Vb", "Vrb", "Verb", "Verbo", "Verbos", "Verbose"],
        ["D", "De", "Dbg", "Dbug", "Debug"],
        ["I", "In", "Inf", "Info", "Infor", "Inform", "Informa", "Informat", "Informati", "Informatio", "Information"],
        ["W", "Wn", "Wrn", "Warn", "Warni", "Warnin", "Warning"],
        ["E", "Er", "Err", "Eror", "Error"],
        ["F", "Fa", "Ftl", "Fatl", "Fatal"]
    ]

    _lowerCaseLevelMap = [
        ["v", "vb", "vrb", "verb", "verbo", "verbos", "verbose"],
        ["d", "de", "dbg", "dbug", "debug"],
        ["i", "in", "inf", "info", "infor", "inform", "informa", "informat", "informati", "informatio", "information"],
        ["w", "wn", "wrn", "warn", "warni", "warnin", "warning"],
        ["e", "er", "err", "eror", "error"],
        ["f", "fa", "ftl", "fatl", "fatal"]
    ]

    _upperCaseLevelMap = [
        ["V", "VB", "VRB", "VERB", "VERBO", "VERBOS", "VERBOSE"],
        ["D", "DE", "DBG", "DBUG", "DEBUG"],
        ["I", "IN", "INF", "INFO", "INFOR", "INFORM", "INFORMA", "INFORMAT", "INFORMATI", "INFORMATIO", "INFORMATION"],
        ["W", "WN", "WRN", "WARN", "WARNI", "WARNIN", "WARNING"],
        ["E", "ER", "ERR", "EROR", "ERROR"],
        ["F", "FA", "FTL", "FATL", "FATAL"]
    ]

    @staticmethod
    def get_level_moniker(value: LogEventLevel, text_format: str | None = None):
        if value.get_int() < LevelAlias.minimum.get_int() or value.get_int() > LevelAlias.maximum.get_int():
            return Casing.format(str(value), text_format)

        if text_format is None or len(text_format) != 2 and len(text_format) != 3:
            val = LevelOutputFormat.__get_level_moniker_from_list(LevelOutputFormat._titleCaseLevelMap, value)
            return Casing.format(val, text_format)

        width = int(text_format[1:])
        if width < 1:
            return ""

        match text_format[0]:
            case "w":
                return LevelOutputFormat.__get_level_moniker_from_list(LevelOutputFormat._lowerCaseLevelMap, value,
                                                                       width)
            case "u":
                return LevelOutputFormat.__get_level_moniker_from_list(LevelOutputFormat._upperCaseLevelMap, value,
                                                                       width)
            case "t":
                return LevelOutputFormat.__get_level_moniker_from_list(LevelOutputFormat._titleCaseLevelMap, value,
                                                                       width)
            case _:
                val = LevelOutputFormat.__get_level_moniker_from_list(LevelOutputFormat._titleCaseLevelMap, value)
                return Casing.format(val, text_format)

    @staticmethod
    def __get_level_moniker_from_list(case_level_map: list[list[str]], level: LogEventLevel, width: int = None):
        case_level = case_level_map[level.get_int()]
        index = len(case_level)
        if width is not None and width < index:
            index = width

        return case_level[index-1]



from pyserilog.core.string_writable import StringWriteable
from typing import Callable

from pyserilog.events.log_event_property_value import LogEventPropertyValue


class SequenceValue(LogEventPropertyValue):
    def __init__(self, elements: list[LogEventPropertyValue]):
        self._elements = elements

    @staticmethod
    def empty():
        return SequenceValue([])

    @property
    def elements(self) -> list[LogEventPropertyValue]:
        return self._elements

    def render(self, output: StringWriteable, text_format: str | None = None, formatter: Callable[[object], str] = None):
        output.write("[")
        for elem in self.elements[:-1]:
            elem.render(output)
            output.write(", ")

        if len(self.elements) > 0:
            self._elements[-1].render(output)
        output.write(']')

    def __eq__(self, other):
        if isinstance(other, SequenceValue):
            return other.elements == self.elements
        return False



from pyserilog.core.ilog_event_sink import ILogEventSink
from pyserilog.core.ilog_event_filter import ILogEventFilter
from pyserilog.debuging.self_log import SelfLog
from pyserilog.events.log_event import LogEvent


class FilteringSink(ILogEventSink):
    def __init__(self, sink: ILogEventSink, filters: list[ILogEventFilter], propagate_exceptions: bool):
        self._sink = sink
        self._filters = filters
        self._propagate_exceptions = propagate_exceptions

    def emit(self, log_event: LogEvent):
        try:
            for log_event_filter in self._filters:
                if not log_event_filter.is_enable(log_event):
                    return
            self._sink.emit(log_event)
        except Exception as ex:
            SelfLog.write_line("Caught exception while applying filters: {0}", ex)
            if self._propagate_exceptions:
                raise ex



from pyserilog.guard import Guard
from pyserilog.core.ilog_event_sink import ILogEventSink
from pyserilog.debuging.self_log import SelfLog
from pyserilog.events.log_event import LogEvent


class SafeAggregateSink(ILogEventSink):
    def __init__(self, sinks: list[ILogEventSink]):
        Guard.against_null(sinks)
        self._sinks = sinks

    def emit(self, log_event: LogEvent):
        for sink in self._sinks:
            try:
                sink.emit(log_event)
            except Exception as ex:
                SelfLog.write_line ("Caught exception while emitting to sink {0}: {1}", sink, ex)


from abc import ABC, abstractmethod

from pyserilog.events.log_event_property_value import LogEventPropertyValue


class ILogEventPropertyValueFactory(ABC):

    @abstractmethod
    def create_property_value(self, value, destructure_objects: bool = False) -> LogEventPropertyValue:
        pass



from abc import ABC, abstractmethod
from pyserilog.core.string_writable import StringWriteable


class MessageTemplateToken(ABC):

    def __init__(self, start_index: int):
        self._start_index = start_index

    @property
    def start_index(self) -> int:
        return self._start_index

    @abstractmethod
    def length(self) -> int:
        pass

    @abstractmethod
    def render(self, properties: dict, writer: StringWriteable):
        pass



from pyserilog import Guard
from pyserilog.core.ilog_event_enricher import ILogEventEnricher
from pyserilog.core.ilog_event_property_factory import ILogEventPropertyFactory
from pyserilog.events import property_utils
from pyserilog.events.log_event import LogEvent


class PropertyEnricher(ILogEventEnricher):
    def __init__(self, name: str, value, destructure_objects: bool = False):
        property_utils.ensure_valid_name(name)

        self._name = name
        self._value = value
        self._destructure_objects = destructure_objects

    def enrich(self, log_event: LogEvent, property_factory: ILogEventPropertyFactory):
        Guard.against_null(log_event)
        Guard.against_null(property_factory)

        property_value = property_factory.create_property(self._name, self._value, self._destructure_objects)
        log_event.add_property_if_absent(property_value)



from typing import Callable

from pyserilog.guard import Guard
import pyserilog.logger_configuration as logConf
from pyserilog.core.logging_level_switch import LoggingLevelSwitch
from pyserilog.core.ilog_event_sink import ILogEventSink
from pyserilog.core.sinks.conditional_sink import ConditionalSink
from pyserilog.core.sinks.dispose_delegating_sink import DisposeDelegatingSink
from pyserilog.core.sinks.disposing_aggregate_sink import DisposingAggregateSink
from pyserilog.core.sinks.secondary_logger_sink import SecondaryLoggerSink
from pyserilog.core.sinks.restricted_sink import RestrictedSink
from pyserilog.debuging.self_log import SelfLog
from pyserilog.events.log_event import LogEvent
from pyserilog.events.log_event_level import LogEventLevel
from pyserilog.events.level_alias import LevelAlias


class LoggerSinkConfiguration:

    def __init__(self, logger_configuration, add_sink_action):
        self._logger_configuration = Guard.against_null(logger_configuration)
        self._add_sink_action = Guard.against_null(add_sink_action)

    def sink(self, log_event_sink: ILogEventSink, restricted_to_minimum_level: LogEventLevel = LevelAlias.minimum,
             level_switch: LoggingLevelSwitch | None = None):
        sink = log_event_sink
        if level_switch is not None:
            if restricted_to_minimum_level != LevelAlias.minimum:
                SelfLog.write_line(
                    "Sink {0} was configured with both a level switch and minimum level '{1}'; " + \
                    "the minimum level will be ignored and the switch level used",
                    sink, restricted_to_minimum_level)

            sink = RestrictedSink(log_event_sink, level_switch)
        elif restricted_to_minimum_level > LevelAlias.minimum:
            sink = RestrictedSink(log_event_sink, LoggingLevelSwitch(restricted_to_minimum_level))

        self._add_sink_action(sink)
        return self._logger_configuration

    def logger(self, configure_logger,
               restricted_to_minimum_level: LogEventLevel = LevelAlias.minimum,
               level_switch: LoggingLevelSwitch | None = None):
        Guard.against_null(configure_logger)

        from pyserilog.core.logger import Logger
        if isinstance(configure_logger, Logger):
            if configure_logger.has_override_map:
                SelfLog.write_line("Minimum level overrides are not supported on sub-loggers " +
                                   "and may be removed completely in a future version.")
            second_sink = SecondaryLoggerSink(configure_logger, attempt_dispose=False)
            return self.sink(second_sink, restricted_to_minimum_level)
        log_configuration = logConf.LoggerConfiguration().minimum_level.level_is(LevelAlias.minimum)
        configure_logger(log_configuration)

        sub_logger = log_configuration.create_logger()

        if sub_logger.has_override_map:
            SelfLog.write_line(
                "Minimum level overrides are not supported on sub-loggers and may be removed" + \
                " completely in a future version.")

        secondary_sink = SecondaryLoggerSink(sub_logger, attempt_dispose=True)
        return self.sink(secondary_sink, restricted_to_minimum_level, level_switch)

    def conditional(self, condition: Callable[[LogEvent], bool], configure_sink: Callable):
        Guard.against_null(condition)
        Guard.against_null(configure_sink)

        return self.wrap(self, lambda s: ConditionalSink(s, condition), configure_sink, LevelAlias.minimum, None)

    @staticmethod
    def wrap(logger_sink_configuration, wrap_sink: Callable[[ILogEventSink], ILogEventSink],
             configure_wrapped_sink: Callable,
             restricted_to_minimum_level=LevelAlias.minimum, level_switch: LoggingLevelSwitch = None):
        """
        Helper method for wrapping sinks.
        :param logger_sink_configuration: The parent sink configuration.
        :param wrap_sink:
        :param configure_wrapped_sink:
        :param restricted_to_minimum_level: The minimum level for events passed through the sink.
         Ignored when "levelSwitch" is specified.
        :param level_switch: A switch allowing the pass-through minimum level to be changed at runtime. Can be None
        :return: Configuration object allowing method chaining
        """
        Guard.against_null(logger_sink_configuration)
        Guard.against_null(wrap_sink)
        Guard.against_null(configure_wrapped_sink)

        sinks_to_wrap: list[ILogEventSink] = []
        from pyserilog.logger_configuration import LoggerConfiguration
        capturing_configuration = LoggerConfiguration()
        capturing_logger_sink_configuration = LoggerSinkConfiguration(capturing_configuration, sinks_to_wrap.append)

        capturing_configuration._write_to = capturing_logger_sink_configuration

        configure_wrapped_sink(capturing_logger_sink_configuration)

        if len(sinks_to_wrap) == 0:
            return logger_sink_configuration._logger_configuration

        enclosed = sinks_to_wrap[0] if len(sinks_to_wrap) == 1 else DisposingAggregateSink(sinks_to_wrap)

        wrapper = wrap_sink(enclosed)

        if not hasattr(wrapper, "__exit__") and hasattr(enclosed, "__exit__"):
            wrapper = DisposeDelegatingSink(wrapper, enclosed)
        return logger_sink_configuration.sink(wrapper, restricted_to_minimum_level, level_switch)



from enum import Enum


class AlignmentDirection(Enum):
    """
    Defines the direction of the alignment.
    """
    LEFT = 0
    """
    Text will be left-aligned.
    """
    RIGHT = 1
    """
    Text will be right-aligned.
    """


class Alignment:
    """
    A structure representing the alignment settings to apply when rendering a property.
    """

    def __init__(self, direction: AlignmentDirection, width: int):
        self.__direction = direction
        self.__width = width

    @property
    def direction(self) -> AlignmentDirection:
        return self.__direction

    @property
    def width(self) -> int:
        return self.__width

    def __eq__(self, other):
        if not isinstance(other, Alignment):
            return False
        return self.__width == other.width and self.direction == other.direction



from datetime import datetime

from pyserilog.core.string_writable import StringWriteable

from pyserilog.guard import Guard
from pyserilog.events.log_event_level import LogEventLevel
from pyserilog.events.log_event_property_value import LogEventPropertyValue
from pyserilog.events.event_property import EventProperty
from pyserilog.events.log_event_property import LogEventProperty
from pyserilog.events.message_template import  MessageTemplate


class LogEvent:

    def __init__(self, timestamp: datetime, level: LogEventLevel, exception: Exception | None,
                 message_template: MessageTemplate,
                 properties: dict[str, LogEventPropertyValue] | list[LogEventProperty]):
        self._timestamp = timestamp
        self._level = level
        self._exception = exception
        self._message_template = message_template
        self._properties: dict[str, LogEventPropertyValue] = {}
        if isinstance(properties, dict):
            self._properties = properties
        elif isinstance(properties, list):
            for prop in properties:
                self.add_or_update_property(prop)

    @property
    def timestamp(self) -> datetime:
        """
        The time at which the event occurred.
        """
        return self._timestamp

    @property
    def level(self) -> LogEventLevel:
        """
        The level of the event.
        """
        return self._level

    @property
    def message_template(self) -> MessageTemplate:
        """
        The message template describing the event.
        """
        return self._message_template

    @property
    def properties(self) -> dict[str, LogEventPropertyValue]:
        return self._properties

    @property
    def exception(self) -> Exception | None:
        """
        An exception associated with the event, or null
        """
        return self._exception

    def render_message(self, output: StringWriteable | None = None):
        """
        Render the message template to the specified output, given the properties associated with the event.
        """
        if output is None:
            return self._message_template.render(properties=self._properties)
        else:
            return self._message_template.render(properties=self._properties, output=output)

    def add_or_update_property(self, prop: LogEventProperty | EventProperty):
        """
        Add a property to the event if not already present, otherwise, update its value.
        :param prop: The property to add or update.
        """
        Guard.against_null(prop)

        self._properties[prop.name] = prop.value

    def add_property_if_absent(self, prop: LogEventProperty | EventProperty):
        Guard.against_null(prop)

        if prop.name not in self._properties:
            self._properties[prop.name] = prop.value

    def remove_property_if_present(self, property_name: str):
        """
        Remove a property from the event, if present. Otherwise no action is performed.
        """
        self._properties.pop(property_name)

    def copy(self):
        properties = {}
        for property_name in self._properties:
            properties[property_name] = self._properties[property_name]

        return LogEvent(timestamp=self.timestamp, level=self.level, exception=self.exception,
                        message_template=self.message_template, properties=properties)



from abc import ABC, abstractmethod

from pyserilog.core.string_writable import StringWriteable


class ITextFormatter(ABC):

    @abstractmethod
    def format(self, log_event, output: StringWriteable):
        pass



class OutputProperties:
    MESSAGE_PROPERTY_NAME: str = "message"
    TIMESTAMP_PROPERTY_NAME: str = "timestamp"
    LEVEL_PROPERTY_NAME: str = "level"
    NEW_LINE_PROPERTY_NAME: str = "new_line"
    EXCEPTION_PROPERTY_NAME: str = "exception"
    PROPERTIES_PROPERTY_NAME: str = "properties"



from typing import Callable
from pyserilog.core.string_writable import StringWriteable


class SelfLog:
    _output: Callable[[str], None] = None

    @staticmethod
    def write_line(log_format: str, *args):
        if SelfLog._output is not None:
            res = log_format.format(*args)
            SelfLog._output(res)

    @staticmethod
    def enable(output: Callable[[str], None] | StringWriteable):
        if isinstance(output, StringWriteable):
            SelfLog._output = lambda x: SelfLog.__write_with_string_io(x, output)
        else:
            SelfLog._output = output

    @staticmethod
    def disable():
        SelfLog._output = None

    @staticmethod
    def __write_with_string_io(value: str, output: StringWriteable):
        output.write(value)



import datetime
from abc import ABC, abstractmethod

from pyserilog.core.ilog_event_enricher import ILogEventEnricher
from pyserilog.events.log_event import LogEvent
from pyserilog.events.log_event_level import LogEventLevel
from pyserilog.events.log_event_property import LogEventProperty
from pyserilog.events.message_template import MessageTemplate

no_property_values = []


class ILogger(ABC):

    @abstractmethod
    def write_event(self, log_event: LogEvent):
        pass

    def write(self, level: LogEventLevel, message_template: str, property_values: list | None = None,
              exception: Exception | None = None):
        if not self.is_enable(level) or message_template is None:
            return

        (checked, parsed_template, bound_properties) = self.bind_message_template(message_template, property_values)
        if checked:
            event = LogEvent(datetime.datetime.now(), level, exception, parsed_template, bound_properties)
            self.write_event(event)

    def is_enable(self, level: LogEventLevel) -> bool:
        return True

    def _write_with_level(self, level: LogEventLevel, message_template: str, *property_value):
        res = []
        if property_value is not None:
            for p in property_value:
                res.append(p)
        self.write(level, message_template, res)

    def debug(self, message_template: str, *property_value):
        self._write_with_level(LogEventLevel.DEBUG, message_template, *property_value)

    def information(self, message_template: str, *property_value):
        self._write_with_level(LogEventLevel.INFORMATION, message_template, *property_value)

    def warning(self, message_template: str, *property_value):
        self._write_with_level(LogEventLevel.WARNING, message_template, *property_value)

    def error(self, message_template: str, *property_value):
        self._write_with_level(LogEventLevel.ERROR, message_template, *property_value)

    def fatal(self, message_template: str, *property_value):
        self._write_with_level(LogEventLevel.FATAL, message_template, *property_value)

    @abstractmethod
    def bind_message_template(self, message_template: str, property_values: list[object] | None) -> \
            tuple[bool, MessageTemplate, list[LogEventProperty]]:
        pass

    def for_context(self, **kwargs):
        """
        enricher: ILogEventEnricher
        :param kwargs:
        :return:
        """

        def get_value(key, default_value=None):
            if key in kwargs:
                return kwargs[key]
            return default_value

        if len(kwargs) == 1:
            if 'enricher' in kwargs:
                enricher = get_value('enricher')
                if isinstance(enricher, ILogEventEnricher):
                    return self._for_enricher_context(enricher)
                else:
                    raise TypeError("enricher should be type of ILogEventEnricher")
            if 'class_type' in kwargs:
                class_type = get_value('class_type')
                if isinstance(class_type, type):
                    return self._for_class_type_context(class_type)
                else:
                    raise TypeError("enricher should be type of ILogEventEnricher")
            elif 'enrichers' in kwargs:
                enrichers = get_value('enrichers')
                raise NotImplementedError
        elif (len(kwargs) == 2 or len(kwargs) == 3) and 'property_name' in kwargs:
            property_name: str = kwargs['property_name']
            value = get_value('value')
            destructure_objects = get_value('destructureObjects', False)
            return self._for_property_name_context(property_name, value, destructure_objects)
        raise NotImplementedError

    @abstractmethod
    def _for_enricher_context(self, enricher: ILogEventEnricher):
        pass

    @abstractmethod
    def _for_property_name_context(self, property_name: str, value, destructure_objects: bool = False):
        pass

    @abstractmethod
    def _for_class_type_context(self, class_type: type):
        pass

    @abstractmethod
    def bind_property(self, property_name: str, value, destructure_objects) -> tuple[bool, LogEventProperty | None]:
        pass



from pyserilog.core.string_writable import StringWriteable
from typing import Callable

from pyserilog.events.log_event_property_value import LogEventPropertyValue
from pyserilog.events.scalar_value import ScalarValue


class DictionaryValue(LogEventPropertyValue):
    def __init__(self, elements: list[tuple[ScalarValue, LogEventPropertyValue]]):
        self._elements = elements

    @property
    def elements(self) -> list[tuple[ScalarValue, LogEventPropertyValue]]:
        return self._elements

    def render(self, output: StringWriteable, text_format: str | None = None, formatter: Callable[[object], str] = None):
        output.write('[')
        delim: str = '('
        for kvp in self._elements:
            output.write(delim)
            delim = ', ('
            key: ScalarValue = kvp[0]
            key.render(output)
            value: LogEventPropertyValue = kvp[1]
            output.write(": ")
            value.render(output)
            output.write(")")

        output.write(']')



from pyserilog.events.log_event_level import LogEventLevel


class LoggingLevelSwitch:

    def __init__(self, initial_minimum_level: LogEventLevel = LogEventLevel.INFORMATION):
        self._minimum_level = initial_minimum_level

    @property
    def minimum_level(self) -> LogEventLevel:
        return self._minimum_level

    @minimum_level.setter
    def minimum_level(self, value: LogEventLevel):
        self._minimum_level = value

    def __str__(self):
        return str(self._minimum_level)


from abc import ABC, abstractmethod

from pyserilog.events.log_event_property_value import LogEventPropertyValue
from pyserilog.parsing.destructuring import Destructuring


class IPropertyValueConverter(ABC):
    @abstractmethod
    def create_property_value_check_destructuring_and_depth(self,
                                                            value, destructure_objects: bool,
                                                            depth: int) -> LogEventPropertyValue:
        pass

    @abstractmethod
    def create_property_value_with_destructuring_and_depth(self, value, destructuring: Destructuring,
                                                           depth: int) -> LogEventPropertyValue:
        pass



from typing import Callable

import pyserilog.logger_configuration as logConf
from pyserilog.guard import Guard, NoneError
from pyserilog.core.idestructuring_policy import IDestructuringPolicy
from pyserilog.policies.ProjectedDestructuringPolicy import ProjectedDestructuringPolicy


class LoggerDestructuringConfiguration:
    """
    Controls template parameter destructuring configuration.
    """

    def __init__(self, logger_configuration, add_scalar_func: Callable[[type], None],
                 add_policy_func: Callable[[IDestructuringPolicy], None],
                 set_maximum_depth_func: Callable[[int], None], set_maximum_string_length: Callable[[int], None],
                 set_maximum_collection_count: Callable[[int], None]
                 ):
        self._logger_configuration = Guard.against_null(logger_configuration)
        self._add_scalar_func: Callable[[type], None] = Guard.against_null(add_scalar_func)
        self._add_policy_func: Callable[[IDestructuringPolicy], None] = Guard.against_null(add_policy_func)
        self._set_maximum_depth_func: Callable[[int], None] = Guard.against_null(set_maximum_depth_func)
        self._set_maximum_string_length: Callable[[int], None] = Guard.against_null(set_maximum_string_length)
        self._set_maximum_collection_count: Callable[[int], None] = Guard.against_null(set_maximum_collection_count)

    def as_scalar(self, scalar_type: type):
        """
        Treat objects of the specified type as scalar values, i.e., don't
        break them down into properties even when destructuring complex types.
        :param scalar_type: Type to treat as scalar.
        :return: Configuration object allowing method chaining.
        """
        Guard.against_null(scalar_type)

        self._add_scalar_func(scalar_type)
        return self._logger_configuration

    def with_policies(self, *destructuring_policies):
        """
        When destructuring objects, transform instances with the provided policies.
        :param destructuring_policies: Policies to apply when destructuring.
        :return: Configuration object allowing method chaining.
        """

        for destructuring_policy in destructuring_policies:
            if destructuring_policy is None:
                return NoneError("None policy is not allowed.")
            if isinstance(destructuring_policy, IDestructuringPolicy):
                self._add_policy_func(destructuring_policy)
            else:
                raise ValueError("type of policies should be IDestructuringPolicy")

        return self._logger_configuration

    def by_transforming(self, value_type: type, transformation_func: Callable):
        """
        When destructuring objects, transform instances of the specified type with the provided function.
        :param value_type:
        :param transformation_func:
        :return:
        """
        Guard.against_null(transformation_func)

        policy = ProjectedDestructuringPolicy(can_apply_func=lambda x: x == value_type,
                                              projection_func=lambda o: transformation_func(o))
        return self.with_policies(policy)

    def to_maximum_depth(self, maximum_destructuring_depth: int):
        """
        When destructuring objects, depth will be limited to 10 property traversals deep to
        guard against ballooning space when recursive/cyclic structures are accidentally passed. To
        change this limit pass a new maximum depth.
        :param maximum_destructuring_depth: The maximum depth to use.
        :return: Configuration object allowing method chaining.
        """
        if maximum_destructuring_depth < 0:
            raise ValueError("Maximum destructuring depth must be positive.")

        self._set_maximum_depth_func(maximum_destructuring_depth)
        return self._logger_configuration

    def to_maximum_string_length(self, maximum_string_length: int):
        """
        When destructuring objects, string values can be restricted to specified length
        thus avoiding bloating payload. Limit is applied to each value separately,
        sum of length of strings can exceed limit.
        :param maximum_string_length: The maximum string length.
        :return: Configuration object allowing method chaining.
        """
        if maximum_string_length < 2:
            raise ValueError("Maximum string length must be at least two.")

        self._set_maximum_string_length(maximum_string_length)
        return self._logger_configuration

    def to_maximum_collection_count(self, maximum_collection_count: int):
        """
        When destructuring objects, collections be restricted to specified count
        thus avoiding bloating payload. Limit is applied to each collection separately,
        sum of length of collection can exceed limit.
        Applies limit to all "list" including dictionaries.
        :param maximum_collection_count:
        :return: Configuration object allowing method chaining.
        """
        if maximum_collection_count < 1:
            raise ValueError("Maximum collection length must be at least one.")
        self._set_maximum_collection_count(maximum_collection_count)
        return self._logger_configuration



from pyserilog import ILogger, Guard
from pyserilog.core.ilog_event_sink import ILogEventSink
from pyserilog.events.log_event import LogEvent


class SecondaryLoggerSink(ILogEventSink):
    def __init__(self, logger: ILogger, attempt_dispose: bool = False):
        self._logger: ILogger = Guard.against_null(logger)
        self._attempt_dispose: bool = attempt_dispose

    def emit(self, log_event: LogEvent):
        Guard.against_null(log_event)

        copy = log_event.copy()
        self._logger.write_event(copy)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self._attempt_dispose:
            return

        if hasattr(self._logger, "__exit__"):
            self._logger.__exit__(exc_type, exc_val, exc_tb)



from pyserilog.events.log_event import LogEvent
from pyserilog.guard import Guard
from pyserilog.core import logging_level_switch
from pyserilog.core.ilog_event_sink import ILogEventSink


class RestrictedSink(ILogEventSink):

    def __init__(self, sink: ILogEventSink, level_switch: logging_level_switch):
        self._sink: ILogEventSink = Guard.against_null(sink)
        self._levelSwitch: logging_level_switch = Guard.against_null(level_switch)

    def emit(self, log_event: LogEvent):
        Guard.against_null(log_event)

        if log_event.level < self._levelSwitch.minimum_level:
            return
        self._sink.emit(log_event)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if hasattr(self._sink, "__exit__"):
            self._sink.__exit__(exc_type, exc_val, exc_tb)



from typing import TypeVar, Generic, Mapping, MappingView

from pyserilog.guard import Guard
from pyserilog.data.log_event_property_value_visitor import LogEventPropertyValueVisitor, TResult
from pyserilog.events.dictionary_value import DictionaryValue
from pyserilog.events.log_event_property import LogEventProperty
from pyserilog.events.log_event_property_value import LogEventPropertyValue
from pyserilog.events.scalar_value import ScalarValue
from pyserilog.events.sequence_value import SequenceValue
from pyserilog.events.structure_value import StructureValue

TState = TypeVar('TState')


class LogEventPropertyValueRewriter(LogEventPropertyValueVisitor[Generic[TState], LogEventPropertyValue]):
    def _visit_scalar_value(self, state: TState, value: ScalarValue) -> LogEventPropertyValue:
        Guard.against_null(value)
        return value

    def _visit_sequence_value(self, state: TState, sequence: SequenceValue) -> LogEventPropertyValue:
        Guard.against_null(sequence)
        for i in range(len(sequence.elements)):
            original: LogEventPropertyValue = sequence.elements[i]
            if not original == self.visit(state, original):
                consts: list[LogEventPropertyValue] = []
                for j in range(i):
                    consts.append(sequence.elements[j])

                for k in range(i, len(sequence.elements)):
                    vp = self.visit(state, sequence.elements[k])
                    consts.append(vp)
                return SequenceValue(consts)
        return sequence

    def _visit_structure_value(self, state: TState, structure: StructureValue) -> LogEventPropertyValue:
        Guard.against_null(structure)

        for i in range(len(structure.properties)):
            original: LogEventProperty = structure.properties[i]
            if not original.value == self.visit(state, original.value):
                properties: list[LogEventProperty] = []
                for j in range(i):
                    properties.append(structure.properties[j])

                for k in range(i, len(structure.properties)):
                    prop = structure.properties[k]
                    event_property = LogEventProperty(prop.name, self.visit(state, prop.value))
                    properties.append(event_property)
                return StructureValue(properties)
        return structure

    def _visit_dictionary_value(self, state: TState, dictionary: DictionaryValue) -> LogEventPropertyValue:
        Guard.against_null(dictionary)

        for (key, value) in dictionary.elements:
            if value != self.visit(state, value):
                elements = []
                for element in dictionary.elements:
                    visited = self.visit(state, element[1])
                    res = tuple([element[0], visited])
                    elements.append(res)
                return DictionaryValue(elements)
        return dictionary

    def _visit_unsupported_value(self, state: TState, value: DictionaryValue) -> TResult:
        return value



from typing import Callable

from pyserilog.guard import Guard
from pyserilog.core.idestructuring_policy import IDestructuringPolicy
from pyserilog.core.ilog_event_property_value_factory import ILogEventPropertyValueFactory
from pyserilog.events.log_event_property_value import LogEventPropertyValue


class ProjectedDestructuringPolicy(IDestructuringPolicy):
    def __init__(self, can_apply_func: Callable[[type], bool], projection_func: Callable[[object], object]):
        self._can_apply_func: Callable[[type], bool] = Guard.against_null(can_apply_func)
        self._projection_func: Callable[[object], object] = Guard.against_null(projection_func)

    def try_destructure(self, value, property_value_factory: ILogEventPropertyValueFactory) -> \
            tuple[bool, LogEventPropertyValue | None]:
        Guard.against_null(value)

        if not self._can_apply_func(type(value)):
            return False, None
        projected = self._projection_func(value)
        result = property_value_factory.create_property_value(value=projected, destructure_objects=True)
        return True, result



from abc import ABC, abstractmethod
from io import StringIO


class StringWriteable(ABC):

    @abstractmethod
    def write(self, value: str):
        pass

    @abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def __enter__(self):
        return self


class StringIOWriteable(StringWriteable):
    def __init__(self):
        self._output = StringIO()

    def write(self, value: str):
        if isinstance(value, str):
            self._output.write(value)
        else:
            raise ValueError("value type should be string")

    def getvalue(self):
        return self._output.getvalue()

    def close(self):
        self._output.close()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._output.close()

    def __str__(self):
        return self._output.getvalue()



from pyserilog.core.string_writable import StringWriteable

from pyserilog.guard import Guard
from pyserilog.parsing.message_template_token import MessageTemplateToken


class TextToken(MessageTemplateToken):

    def __init__(self, text: str, start_index: int = -1):
        super().__init__(start_index)
        Guard.against_null(text)
        self._text = text

    @property
    def text(self):
        return self._text

    def render(self, properties: dict, writer: StringWriteable):
        pass

    @property
    def length(self) -> int:
        return len(self._text)

    def __eq__(self, other):
        if not isinstance(other, TextToken):
            return False
        return other._text == self._text

    def __str__(self):
        return self._text



from abc import ABC, abstractmethod

from pyserilog.events.log_event import LogEvent


class ILogEventFilter(ABC):
    """
    Provides filtering of the log event stream.
    """

    @abstractmethod
    def is_enable(self, log_event: LogEvent) -> bool:
        """
        Returns true if the provided event is enabled. Otherwise
        :param log_event: The event to test.
        :return: True if the event is enabled by this filter. If False is returned, the event will not be emitted.
        """
        pass



from typing import Callable

from pyserilog.core.constants import Constants
from pyserilog.debuging import SelfLog
from pyserilog.events.log_event import LogEvent
from pyserilog.events.scalar_value import ScalarValue
import pyserilog.utils as utils


class Matching:

    @staticmethod
    def from_source(source: type | str):
        if isinstance(source, type):
            return Matching.with_property(Constants.SOURCE_CONTEXT_PROPERTY_NAME, utils.full_name(source))

        def func(s: str):
            return s is not None and s.startswith(source) and (len(s) == len(source) or s[len(source)] == '.')

        return Matching.with_property(Constants.SOURCE_CONTEXT_PROPERTY_NAME, func)

    @staticmethod
    def with_property(property_name: str, value=None) -> Callable[[LogEvent], bool]:
        def func(x: LogEvent):
            return property_name in x.properties

        if value is None:
            return func

        if isinstance(value, Callable) or isinstance(value, type(lambda: 0)):
            def func_callable(x: LogEvent):
                if property_name not in x.properties:
                    return False
                val = x.properties[property_name]
                if isinstance(val, ScalarValue):
                    try:
                        res = value(val.value)
                        return res
                    except TypeError as ex:
                        return False
                return False

            return func_callable

        def func2(x: LogEvent):
            scalar = ScalarValue(value)
            return property_name in x.properties and scalar == x.properties[property_name]

        return func2



from pyserilog.core.string_writable import StringWriteable

from pyserilog.parsing.alignment import Alignment
from pyserilog.parsing.destructuring import Destructuring
from pyserilog.parsing.message_template_token import MessageTemplateToken


class PropertyTokenBase(MessageTemplateToken):

    def __init__(self, property_name: str, raw_text: str, text_format: str = None,
                 text_alignment: Alignment = None, destructuring: Destructuring = Destructuring.DEFAULT,
                 start_index: int = -1):
        super().__init__(start_index)
        self._property_name = property_name
        self._raw_text = raw_text
        self._destructuring = destructuring
        self._text_alignment = text_alignment
        self._text_format = text_format

        self._position = None
        if property_name.isdigit() and int(property_name) >= 0:
            self._position = int(property_name)

    @property
    def length(self) -> int:
        return len(self._raw_text)

    @property
    def property_name(self) -> str:
        return self._property_name

    @property
    def is_positional(self) -> bool:
        return self._position is not None

    @property
    def destructuring(self) -> Destructuring:
        return self._destructuring

    @property
    def format(self):
        return self._text_format

    @property
    def alignment(self) -> Alignment | None:
        return self._text_alignment

    @property
    def raw_text(self):
        return self._raw_text

    def render(self, properties: dict, writer: StringWriteable):
        pass

    def try_get_positional_value(self) -> tuple[bool, int]:
        if self._position is None:
            return False, 0
        return True, self._position

    def __eq__(self, other):
        if not isinstance(other, PropertyToken):
            return False
        return other._destructuring == self._destructuring and other.format == self.format and \
               other._property_name == self._property_name and other._raw_text == self._raw_text

    def __hash__(self):
        return hash(self._property_name)

    def __str__(self):
        return self.raw_text


class PropertyToken(PropertyTokenBase):
    __metaclass__ = PropertyTokenBase



from pyserilog.capturing.property_value_converter import PropertyValueConverter
from pyserilog.debuging import SelfLog
from pyserilog.events.event_property import EventProperty
from pyserilog.events.message_template import MessageTemplate
from pyserilog.parsing.property_token import PropertyToken


class PropertyBinder:
    def __init__(self, value_converter: PropertyValueConverter):
        self._value_converter = value_converter

    def construct_properties(self, message_template: MessageTemplate, message_template_parameters: list[object]) -> \
            list[EventProperty]:
        if message_template_parameters is None or len(message_template_parameters) == 0:
            if message_template.named_properties is not None or message_template.positional_properties is not None:
                SelfLog.write_line("Required properties not provided for: {0}", message_template)
            return []

        if message_template.positional_properties is not None:
            return self._construct_positional_properties(message_template, message_template_parameters)
        return self._construct_named_properties(message_template, message_template_parameters)

    def _construct_positional_properties(self, template: MessageTemplate, message_template_parameters: list[object]) \
            -> list[EventProperty]:
        if not len(template.positional_properties) == len(message_template_parameters):
            SelfLog.write_line("Positional property count does not match parameter count: {0}", template)

        result = {}
        for prop in template.positional_properties:
            (have_position, position) = prop.try_get_positional_value()
            if have_position:
                if position < 0 or position >= len(message_template_parameters):
                    SelfLog.write_line("Unassigned positional value {0} in: {1}", position, template)
                else:
                    value = message_template_parameters[position]
                    result[position] = self.__construct_property(property_token=prop, value=value)

        keys = list(result.keys())
        keys.sort()
        final_result = []
        for key in keys:
            final_result.append(result[key])
        return final_result

    def _construct_named_properties(self, template: MessageTemplate, message_template_parameters: list[object]):
        named_properties = template.named_properties
        if named_properties is None:
            return []
        matched_run = len(named_properties)
        if not len(named_properties) == len(message_template_parameters):
            matched_run = min(len(named_properties), len(message_template_parameters))
            SelfLog.write_line("Named property count does not match parameter count: {0}", template)

        result = []
        for i in range(matched_run):
            prop = named_properties[i]
            value = message_template_parameters[i]
            result.append(self.__construct_property(property_token=prop, value=value))

        for i in range(matched_run, len(message_template_parameters)):
            value = self._value_converter.create_property_value_check_destructuring(message_template_parameters[i])
            result.append(EventProperty(f"__{i}", value))
        return result

    def __construct_property(self, property_token: PropertyToken, value) -> EventProperty:
        log_event_prop_value = self._value_converter. \
            create_property_value_with_destructuring(value, property_token.destructuring)
        return EventProperty(property_token.property_name, log_event_prop_value)



import decimal
from pyserilog.core.string_writable import StringWriteable

from pyserilog.guard import Guard
from pyserilog.events.dictionary_value import DictionaryValue
from pyserilog.events.scalar_value import ScalarValue, LogEventPropertyValue
from pyserilog.events.log_event_property import LogEventProperty
from pyserilog.events.sequence_value import SequenceValue
from pyserilog.events.structure_value import StructureValue
from pyserilog.formatting.itext_formatter import ITextFormatter
from datetime import datetime

from pyserilog.formatting.json.json_value_formatter import JsonValueFormatter


class JsonFormatter(ITextFormatter):

    def __init__(self, closing_delimiter: str | None = None, render_message: bool = False):
        self._close_delimiter: str = closing_delimiter if closing_delimiter is not None else '\n'
        self._render_message = render_message
        self._literal_writers = {
            int: self.__write_to_string,
            float: self.__write_to_string,
            decimal.Decimal: self.__write_to_string,
            datetime: lambda x, force_quotation, output: self.__write_datetime(x, output),
            ScalarValue: lambda x, quote, output: self.__write_literal(x.value, output, quote),
            SequenceValue: lambda x, quote, output: self.__write_sequence(x.elements, output, quote),
            DictionaryValue: lambda x, quote, output: self.__write_dictionary(x.elements, output),
            StructureValue: lambda x, quote, output: self.__write_structure(x.type_tag, x.properties, output),
        }

    def format(self, log_event, output: StringWriteable):
        from pyserilog.events.log_event import LogEvent
        log_event: LogEvent = log_event
        Guard.against_null(log_event)
        Guard.against_null(output)

        output.write("{")
        delim = JsonFormatter.Delimiter()
        self.__write_timestamp(timestamp=log_event.timestamp, delim=delim, output=output)
        if self._render_message:
            raise NotImplementedError

        if len(log_event.properties) > 0:
            self.__write_properties(props=log_event.properties, output=output)

        output.write("}")

    def __write_timestamp(self, timestamp: datetime, delim, output: StringWriteable):
        self.__write_json_property("timestamp", timestamp, delim, output)

    def __write_json_property(self, name: str, value, preceding_delimiter, output: StringWriteable):
        if preceding_delimiter.delim is not None:
            output.write(preceding_delimiter.delim)
        output.write('"')
        output.write(name)
        output.write('":')
        self.__write_literal(value, output)
        preceding_delimiter.delim = ","

    def __write_literal_value(self, value, output: StringWriteable):
        self.write_string(str(value), output)

    def __write_literal(self, value, output: StringWriteable, force_quotation: bool = False):
        if value is None:
            output.write("null")
            return

        if type(value) in self._literal_writers:
            func = self._literal_writers[type(value)]
            func(value, force_quotation, output)
            return
        self.__write_literal_value(value, output)

    @staticmethod
    def __write_to_string(number, quote: bool, output: StringWriteable):
        if quote:
            output.write('"')
        output.write(str(number))
        if quote:
            output.write('"')

    def __write_properties(self, props: dict[str, LogEventPropertyValue], output: StringWriteable):
        output.write(",\"properties\":{")
        self.__write_properties_values(props=props, output=output)
        output.write("}")

    def __write_properties_values(self, props: dict[str, LogEventPropertyValue], output: StringWriteable):
        preceding_delimiter = JsonFormatter.Delimiter()
        for property_name in props:
            property_value = props[property_name]
            self.__write_json_property(property_name, property_value, preceding_delimiter, output)

    @staticmethod
    def __write_datetime(value: datetime, output: StringWriteable):
        output.write('"')
        res = JsonValueFormatter.datetime_format(value)
        output.write(res)
        output.write('"')

    @staticmethod
    def write_string(value: str, output: StringWriteable):
        JsonValueFormatter.write_quoted_json_string(value, output)

    def __write_sequence(self, elements: list, output: StringWriteable, force_quotation: bool = False):
        output.write("[")
        delim = JsonFormatter.Delimiter()

        for value in elements:
            if delim.delim is not None:
                output.write(delim.delim)
            delim.delim = ','
            self.__write_literal(value=value, output=output, force_quotation=force_quotation)

        output.write("]")

    def __write_dictionary(self, elements: list[tuple[ScalarValue, LogEventPropertyValue]], output: StringWriteable):
        output.write("{")
        delim = JsonFormatter.Delimiter()

        for (key, value) in elements:
            if delim.delim is not None:
                output.write(delim.delim)
            delim.delim = ","
            self.__write_literal(key, output, force_quotation=True)
            output.write(":")
            self.__write_literal(value, output)

        output.write("}")

    def __write_structure(self, type_tag: str, properties: list[LogEventProperty],
                          output: StringWriteable):
        output.write("{")
        delim = JsonFormatter.Delimiter()
        if type_tag is not None:
            self.__write_json_property("_type_tag", type_tag, delim, output)
        for prop in properties:
            self.__write_json_property(prop.name, prop.value, delim, output)
        output.write("}")

    class Delimiter:
        def __init__(self):
            self._delim = None

        @property
        def delim(self):
            return self._delim

        @delim.setter
        def delim(self, value):
            self._delim = value



from typing import Callable

from pyserilog.guard import Guard

from pyserilog.core.ilog_event_enricher import ILogEventEnricher
from pyserilog.core.logging_level_switch import LoggingLevelSwitch
from pyserilog.core.enrichers.conditional_enricher import ConditionalEnricher
from pyserilog.core.enrichers.property_enricher import PropertyEnricher
from pyserilog.core.enrichers.safe_aggregate_enricher import SafeAggregateEnricher
from pyserilog.debuging import SelfLog
from pyserilog.events.log_event import LogEvent
from pyserilog.events.log_event_level import LogEventLevel


class LoggerEnrichmentConfiguration:
    def __init__(self, logger_configuration, add_enricher_function: Callable[[ILogEventEnricher], None]):
        self._logger_configuration = Guard.against_null(logger_configuration)
        self._add_enricher_function = Guard.against_null(add_enricher_function)

    def with_enrichers(self, *enrichers):
        for log_event_enricher in enrichers:
            Guard.against_null(value=log_event_enricher)

            if isinstance(log_event_enricher, ILogEventEnricher):
                self._add_enricher_function(log_event_enricher)
            else:
                raise TypeError(f"enrichers should be type of ILogEventEnricher , error for {log_event_enricher}")
        return self._logger_configuration

    def with_property(self, name: str, value, destructure_objects: bool = False):
        enricher = PropertyEnricher(name, value, destructure_objects)
        return self.with_enrichers(enricher)

    def when(self, condition: Callable[[LogEvent], bool], configure_enricher: Callable):
        Guard.against_null(condition)
        Guard.against_null(configure_enricher)

        return self.wrap(self, lambda e: ConditionalEnricher(e, condition), configure_enricher)

    def at_level(self, level_switch: LoggingLevelSwitch | LogEventLevel, configure_enricher: Callable):
        Guard.against_null(configure_enricher)

        lev: LoggingLevelSwitch = level_switch
        if isinstance(lev, LogEventLevel):
            lev = LoggingLevelSwitch(lev)
        return self.wrap(self, lambda e: ConditionalEnricher(e, lambda le: le.level >= lev.minimum_level),
                         configure_enricher)

    @staticmethod
    def wrap(logger_enrichment_configuration, wrap_enricher: Callable[[ILogEventEnricher], ILogEventEnricher],
             configure_wrapped_enricher: Callable):
        Guard.against_null(logger_enrichment_configuration)
        Guard.against_null(wrap_enricher)
        Guard.against_null(configure_wrapped_enricher)

        enrichers_to_wrap = []
        from pyserilog.logger_configuration import LoggerConfiguration
        capturing_configuration = LoggerConfiguration()
        capturing_logger_enrichment_configuration = LoggerEnrichmentConfiguration(capturing_configuration,
                                                                                  enrichers_to_wrap.append)

        capturing_configuration.enrich = capturing_logger_enrichment_configuration

        configure_wrapped_enricher(capturing_logger_enrichment_configuration)

        if len(enrichers_to_wrap) == 0:
            return logger_enrichment_configuration._logger_configuration

        enclosed = enrichers_to_wrap[0] if len(enrichers_to_wrap) == 0 else SafeAggregateEnricher(enrichers_to_wrap)

        wrapped_enricher = wrap_enricher(enclosed)

        if not hasattr(wrapped_enricher, "__exit__"):
            SelfLog.write_line("Wrapping enricher {0} does not implement IDisposable; to ensure " + \
                               "wrapped enrichers are properly disposed, wrappers should dispose " + \
                               "their wrapped contents", wrapped_enricher)

        return logger_enrichment_configuration.with_enrichers(wrapped_enricher)



from abc import ABC, abstractmethod
from typing import TypeVar, Generic

from pyserilog.guard import Guard
from pyserilog.events.dictionary_value import DictionaryValue
from pyserilog.events.log_event_property_value import LogEventPropertyValue
from pyserilog.events.scalar_value import ScalarValue
from pyserilog.events.sequence_value import SequenceValue
from pyserilog.events.structure_value import StructureValue

TState = TypeVar('TState')
TResult = TypeVar('TResult')


class LogEventPropertyValueVisitor(ABC, Generic[TState, TResult]):

    def visit(self, state: TState, value: LogEventPropertyValue) -> TResult:
        Guard.against_null(value)

        if isinstance(value, ScalarValue):
            return self._visit_scalar_value(state, value)
        if isinstance(value, SequenceValue):
            return self._visit_sequence_value(state, value)
        if isinstance(value, StructureValue):
            return self._visit_structure_value(state, value)
        if isinstance(value, DictionaryValue):
            return self._visit_dictionary_value(state, value)
        return self._visit_unsupported_value(state, value)

    @abstractmethod
    def _visit_scalar_value(self, state: TState, value: ScalarValue) -> TResult:
        pass

    @abstractmethod
    def _visit_sequence_value(self, state: TState, value: SequenceValue) -> TResult:
        pass

    @abstractmethod
    def _visit_structure_value(self, state: TState, value: StructureValue) -> TResult:
        pass

    @abstractmethod
    def _visit_dictionary_value(self, state: TState, value: DictionaryValue) -> TResult:
        pass

    def _visit_unsupported_value(self, state: TState, value: DictionaryValue) -> TResult:
        Guard.against_null(value)
        raise Exception(f"The value {value} is not of a type supported by this visitor.")



from builtins import NotImplementedError

from pyserilog.core.string_writable import StringIOWriteable, StringWriteable

from pyserilog.formatting.display.level_output_format import LevelOutputFormat
from pyserilog.formatting.display.properties_output_format import PropertiesOutputFormat
from pyserilog.guard import Guard
from pyserilog.events.log_event import LogEvent
from pyserilog.events.message_template import MessageTemplate
from pyserilog.events.scalar_value import ScalarValue
from pyserilog.formatting import ITextFormatter
from pyserilog.formatting.display.output_properties import OutputProperties
from pyserilog.parsing.message_template_parser import MessageTemplateParser
from pyserilog.parsing.property_token import PropertyToken
from pyserilog.parsing.text_token import TextToken
from pyserilog.rendering.casing import Casing
from pyserilog.rendering.message_template_renderer import MessageTemplateRenderer
from pyserilog.rendering.padding import Padding


class MessageTemplateTextFormatter(ITextFormatter):
    def __init__(self, output_template: str):
        Guard.against_null(output_template)
        self._output_template: MessageTemplate = MessageTemplateParser().parse(output_template)

    def format(self, log_event: LogEvent, output: StringWriteable):
        for token in self._output_template.tokens:
            if isinstance(token, TextToken):
                MessageTemplateRenderer.render_text_token(token, output)
                continue
            assert isinstance(token, PropertyToken)
            pt = token
            if pt.property_name == OutputProperties.LEVEL_PROPERTY_NAME:
                moniker = LevelOutputFormat.get_level_moniker(log_event.level, pt.format)
                Padding.apply(output, moniker, pt.alignment)
            elif pt.property_name == OutputProperties.NEW_LINE_PROPERTY_NAME:
                Padding.apply(output, '\n', pt.alignment)
            elif pt.property_name == OutputProperties.EXCEPTION_PROPERTY_NAME:
                raise NotImplementedError
            else:

                writer = StringIOWriteable() if pt.alignment is not None else output

                if pt.property_name == OutputProperties.MESSAGE_PROPERTY_NAME:
                    MessageTemplateRenderer.render(log_event.message_template, log_event.properties, writer, pt.format)
                elif pt.property_name == OutputProperties.TIMESTAMP_PROPERTY_NAME:
                    ScalarValue.render_value(log_event.timestamp, writer, pt.format)
                elif pt.property_name == OutputProperties.PROPERTIES_PROPERTY_NAME:
                    PropertiesOutputFormat.render(log_event.message_template, log_event.properties,
                                                  self._output_template, writer, pt.format)
                else:
                    # If a property is missing, don't render anything (message templates render the raw token here).
                    if pt.property_name not in log_event.properties:
                        continue
                    property_value = log_event.properties[pt.property_name]
                    if isinstance(property_value, ScalarValue) and isinstance(property_value.value, str):
                        cased = Casing.format(property_value.value, pt.format)
                        writer.write(cased)
                    else:
                        property_value.render(writer, pt.format)

                if pt.alignment is not None:
                    Padding.apply(output, writer.getvalue(), pt.alignment)

                    writer.close()



from pyserilog.ILogger import ILogger
from pyserilog.core.ilog_event_enricher import ILogEventEnricher
from pyserilog.events.log_event import LogEvent
from pyserilog.events.log_event_level import LogEventLevel
from pyserilog.events.log_event_property import LogEventProperty
from pyserilog.events.message_template import MessageTemplate


class SilentLogger(ILogger):
    def write_event(self, log_event: LogEvent):
        pass

    def bind_message_template(self, message_template: str, property_values: list[object] | None) -> \
            tuple[bool, MessageTemplate, list[LogEventProperty]]:
        return False, None, None

    def _for_enricher_context(self, enricher: ILogEventEnricher):
        return self

    def _for_property_name_context(self, property_name: str, value, destructure_objects: bool = False):
        return self

    def _for_class_type_context(self, class_type: type):
        return self

    def bind_property(self, property_name: str, value, destructure_objects) -> tuple[bool, LogEventProperty | None]:
        return False, None

    def is_enable(self, level: LogEventLevel) -> bool:
        return False

    def write(self, level: LogEventLevel, message_template: str, property_values: list | None = None,
              exception: Exception | None = None):
        super().write(level, message_template, property_values, exception)

    def _write_with_level(self, level: LogEventLevel, message_template: str, *property_value):
        pass

    def debug(self, message_template: str, *property_value):
        pass

    def information(self, message_template: str, *property_value):
        pass

    def warning(self, message_template: str, *property_value):
        pass

    def error(self, message_template: str, *property_value):
        pass

    def fatal(self, message_template: str, *property_value):
        pass

    def for_context(self, **kwargs):
        return self



from typing import Callable

from pyserilog.guard import Guard
from pyserilog.core.logging_level_switch import LoggingLevelSwitch
from pyserilog.events.log_event_level import LogEventLevel


class LoggerMinimumLevelConfiguration:

    def __init__(self, logger_configuration, set_minimum_func: Callable[[LogEventLevel], None],
                 set_level_switch_func: Callable[[LoggingLevelSwitch], None],
                 add_override_func: Callable[[str, LoggingLevelSwitch], None]):
        self._logger_configuration = Guard.against_null(logger_configuration)
        self._set_minimum_func: Callable[[LogEventLevel], None] = set_minimum_func
        self._set_level_switch_func: Callable[[LoggingLevelSwitch], None] = set_level_switch_func
        self._add_override_func: Callable[[str, LoggingLevelSwitch], None] = add_override_func

    def level_is(self, minimum_level: LogEventLevel):
        """
        Sets the minimum level at which events will be passed to sinks.
        :param minimum_level: The minimum level to set.
        :return: Configuration object allowing method chaining.
        """
        self._set_minimum_func(minimum_level)
        return self._logger_configuration

    def controlled_by(self, level_switch: LoggingLevelSwitch):
        """
        Sets the minimum level to be dynamically controlled by the provided switch.
        :param level_switch: The switch.
        :return: Configuration object allowing method chaining.
        """
        Guard.against_null(level_switch)

        self._set_level_switch_func(level_switch)
        return self._logger_configuration

    def verbose(self):
        """
        Anything and everything you might want to know about a running block of code.
        :return: Configuration object allowing method chaining.
        """
        return self.level_is(LogEventLevel.VERBOSE)

    def debug(self):
        """
        Internal system events that aren't necessarily observable from the outside.
        :return: Configuration object allowing method chaining.
        """
        return self.level_is(LogEventLevel.DEBUG)

    def information(self):
        """
        The lifeblood of operational intelligence - things happen.
        :return: Configuration object allowing method chaining.
        """
        return self.level_is(LogEventLevel.INFORMATION)

    def warning(self):
        """
        Functionality is unavailable, invariants are broken  or data is lost.
        :return: Configuration object allowing method chaining.
        """
        return self.level_is(LogEventLevel.WARNING)

    def error(self):
        """
        Functionality is unavailable, invariants are broken or data is lost.
        :return: Configuration object allowing method chaining.
        """
        return self.level_is(LogEventLevel.ERROR)

    def fatal(self):
        """
        If you have a pager, it goes off when one of these occurs.
        :return:
        """
        return self.level_is(LogEventLevel.FATAL)

    def override(self, source: str, level_switch: LoggingLevelSwitch | LogEventLevel):
        Guard.against_null(source)
        Guard.against_null(level_switch)

        if not isinstance(level_switch, LoggingLevelSwitch):
            return self.override(source, LoggingLevelSwitch(level_switch))
        trimmed = source.strip()
        if len(trimmed) == 0:
            raise ValueError(f"A source source must be provided.")

        self._add_override_func(trimmed, level_switch)
        return self._logger_configuration



from typing import Callable

from pyserilog import Guard
from pyserilog.ILogger import ILogger
from pyserilog.capturing.message_template_processor import MessageTemplateProcessor
from pyserilog.core.constants import Constants
from pyserilog.core.logging_level_switch import LoggingLevelSwitch
from pyserilog.core.ilog_event_enricher import ILogEventEnricher
from pyserilog.core.level_override_map import LevelOverrideMap
from pyserilog.core.ilog_event_sink import ILogEventSink
from pyserilog.core.enrichers.fixed_property_enricher import FixedPropertyEnricher
from pyserilog.core.pipeline.silet_logger import SilentLogger
from pyserilog.debuging import SelfLog
from pyserilog.events import property_utils
from pyserilog.events.event_property import EventProperty
from pyserilog.events.log_event import LogEvent
from pyserilog.events.log_event_level import LogEventLevel
from pyserilog.events.log_event_property import LogEventProperty
from pyserilog.events.message_template import MessageTemplate
import pyserilog.utils as utils


class Logger(ILogger, ILogEventSink):
    none = SilentLogger()

    def __init__(self, message_template_processor: MessageTemplateProcessor, minimum_level: LogEventLevel,
                 level_switch: LoggingLevelSwitch, sink: ILogEventSink, enricher: ILogEventEnricher,
                 exit_function: Callable, override_map: LevelOverrideMap):
        self._message_template_processor = message_template_processor
        self._minimum_level = minimum_level
        self._level_switch = level_switch
        self._sink = sink
        self._exit_function = exit_function
        self._override_map = override_map
        self._enricher = enricher

    def _for_class_type_context(self, class_type: type):
        if class_type is None:
            return self

        name = utils.full_name(class_type)
        return self._for_property_name_context(Constants.SOURCE_CONTEXT_PROPERTY_NAME, name)

    def _for_property_name_context(self, property_name: str, value, destructure_objects: bool = False):
        if not property_utils.is_valid_name(property_name):
            SelfLog.write_line("Attempt to call for_context() with invalid property name `{0}` (value: `{1}`)",
                               property_name, value)
            return self

        #     It'd be nice to do the destructuring lazily, but unfortunately `value` may be mutated between
        #     now and the first log event written.
        property_value = self._message_template_processor.create_property_value(value, destructure_objects)
        enricher = FixedPropertyEnricher(EventProperty(property_name, property_value))

        minimum_level: LogEventLevel = None
        level_switch: LoggingLevelSwitch = None

        if self._override_map is not None and property_name == Constants.SOURCE_CONTEXT_PROPERTY_NAME \
                and isinstance(value, str):
            minimum_level, level_switch = self._override_map.get_effective_level(value)
        else:
            minimum_level = self._minimum_level
            level_switch = self._level_switch

        return Logger(self._message_template_processor, minimum_level, level_switch, sink=self, enricher=enricher,
                      exit_function=None, override_map=self._override_map)

    def _for_enricher_context(self, enricher: ILogEventEnricher):
        raise NotImplementedError

    def bind_message_template(self, message_template: str, property_values: list[object] | None = None) -> \
            tuple[bool, MessageTemplate, list[LogEventProperty]]:
        if message_template is None:
            return False, None, None

        parsed_template, bound_event_properties = self._message_template_processor.process(message_template,
                                                                                           property_values)
        bound_properties = [] if len(bound_event_properties) == 0 else list(
            map(lambda x: LogEventProperty.create_with_event_property(x), bound_event_properties))
        return True, parsed_template, bound_properties

    def write_event(self, log_event: LogEvent):
        if log_event is None:
            return
        if not self.is_enable(log_event.level):
            return
        self.__dispatch(log_event)

    def emit(self, log_event: LogEvent):
        Guard.against_null(log_event)

        self.__dispatch(log_event)

    def is_enable(self, level: LogEventLevel) -> bool:
        if level < self._minimum_level:
            return False
        return self._level_switch is None or level >= self._level_switch.minimum_level

    @property
    def has_override_map(self) -> bool:
        return self._override_map is not None

    def bind_property(self, property_name: str, value, destructure_objects) -> tuple[bool, LogEventProperty | None]:
        if not property_utils.is_valid_name(property_name):
            return False, None
        prop = self._message_template_processor.create_property(property_name, value, destructure_objects)
        return True, prop

    def __dispatch(self, log_event: LogEvent):
        try:
            self._enricher.enrich(log_event, self._message_template_processor)
        except Exception as ex:
            SelfLog.write_line("Exception {0} caught while enriching {1} with {2}.", ex, log_event, self._enricher)
        self._sink.emit(log_event)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._exit_function is not None:
            self._exit_function(exc_type, exc_val, exc_tb)



from pyserilog.guard import Guard
from pyserilog.events.log_event_property_value import LogEventPropertyValue
from pyserilog.events.property_utils import ensure_valid_name


class EventProperty:
    def __init__(self, name: str, value: LogEventPropertyValue):
        Guard.against_null(name)
        ensure_valid_name(name)

        self._name = name
        self._value = value

    @property
    def name(self) -> str:
        return self._name

    @property
    def value(self) -> LogEventPropertyValue:
        return self._value

    def __eq__(self, other):
        if not isinstance(other, EventProperty):
            return False

        return self.name == other.name and self.value == other.value





from abc import ABC, abstractmethod


class ILogEventPropertyFactory(ABC):

    @abstractmethod
    def create_property(self, name : str , value, destructure_objects : bool = False):
        """
        Construct a :ref <see cref="LogEventProperty"/> with the specified name and value.
        :param name: The name of the property.
        :param value:The value of the property.
        :param destructure_objects: If 'True', and the value is a non-primitive, non-array type,
        then the value will be converted to a structure; otherwise, unknown types will
        be converted to scalars, which are generally stored as strings.
        :return: Created "LogEventProperty" instance.
        """
        pass



import string
from builtins import staticmethod
from pyserilog.core.string_writable import StringIOWriteable

from pyserilog.guard import Guard
from pyserilog.core.imessage_template_parser import IMessageTemplateParser
from pyserilog.events.message_template import MessageTemplate
from pyserilog.parsing.alignment import Alignment, AlignmentDirection
from pyserilog.parsing.destructuring import Destructuring
from pyserilog.parsing.message_template_token import MessageTemplateToken
from pyserilog.parsing.property_token import PropertyToken
from pyserilog.parsing.text_token import TextToken


class MessageTemplateParser(IMessageTemplateParser):

    def parse(self, message_template: str) -> MessageTemplate:
        Guard.against_null(message_template)
        tokens = self.__tokenize(message_template)
        return MessageTemplate(tokens, message_template)

    @staticmethod
    def __tokenize(message_template: str) -> list[MessageTemplateToken]:
        result: list[MessageTemplateToken] = []
        if len(message_template) == 0:
            result.append(TextToken("", 0))
            return result
        next_index = 0
        while True:
            before_text = next_index
            tt, next_index = MessageTemplateParser.__parse_text_token(next_index, message_template)
            if next_index > before_text:
                result.append(tt)
            if next_index == len(message_template):
                break
            before_pop = next_index
            pt, next_index = MessageTemplateParser.__parse_property_token(next_index, message_template)
            if before_pop < next_index:
                result.append(pt)
            if next_index == len(message_template):
                break
        return result

    @staticmethod
    def __parse_text_token(start_at: int, message_template: str) -> tuple[TextToken, int]:
        first = start_at
        accum = StringIOWriteable()

        while True:
            nc = message_template[start_at]
            if nc == '{':
                if start_at + 1 < len(message_template) and message_template[start_at + 1] == '{':
                    accum.write(nc)
                    start_at += 1
                else:
                    break
            else:
                accum.write(nc)
                if nc == '}':
                    if start_at + 1 < len(message_template) and message_template[start_at + 1] == '}':
                        start_at += 1
            start_at += 1
            if start_at >= len(message_template):
                break
        next_step = start_at
        result = str(accum)
        token = TextToken(result, first)
        return token, next_step

    @staticmethod
    def __parse_property_token(start_at: int, message_template: str) -> tuple[MessageTemplateToken, int]:
        first = start_at
        start_at += 1
        while start_at < len(message_template) and \
                MessageTemplateParser.__is_valid_in_property_tag(message_template[start_at]):
            start_at += 1

        if start_at == len(message_template) or message_template[start_at] != '}':
            next_value = start_at
            token = TextToken(message_template[first: next_value], first)
            return token, next_value
        next_value = start_at + 1
        raw_text = message_template[first: next_value]
        tag_content = raw_text[1: 1 + (next_value - (first + 2))]
        if len(tag_content) == 0:
            return TextToken(raw_text, first), next_value
        checked, property_name_and_destructuring, tag_content_format, tag_content_alignment = \
            MessageTemplateParser.__try_split_tag_content(tag_content)
        if not checked:
            return TextToken(raw_text, first), next_value
        property_name = property_name_and_destructuring
        destructuring = Destructuring.DEFAULT
        if len(property_name) > 0:
            checked, destructuring = MessageTemplateParser.__try_get_destructuring_hint(property_name[0])
            if checked:
                property_name = property_name[1:]
        if len(property_name) == 0:
            return TextToken(raw_text, first), next_value
        for char in property_name:
            if not MessageTemplateParser.__is_valid_in_property_name(char):
                return TextToken(raw_text, first), next_value
        if tag_content_format is not None:
            for char in tag_content_format:
                if not MessageTemplateParser.__is_valid_in_format(char):
                    return TextToken(raw_text, first), next_value
        text_alignment: Alignment = None
        if tag_content_alignment is not None:
            for char in tag_content_alignment:
                if not MessageTemplateParser.__is_valid_in_alignment(char):
                    return TextToken(raw_text, first), next_value

            def last_index_of(text: str, sub: str):
                try:
                    return text.rindex(sub)
                except ValueError:
                    return -1

            last_dash = last_index_of(tag_content_alignment, '-')
            if last_dash > 0:
                return TextToken(raw_text, first), next_value

            width: str = tag_content_alignment if last_dash == -1 else tag_content_alignment[1:]
            if not width.isdigit() or int(width) == 0:
                return TextToken(raw_text, first), next_value
            width: int = int(width)

            direction = AlignmentDirection.RIGHT if last_dash == -1 else AlignmentDirection.LEFT
            text_alignment = Alignment(direction, width)
        return PropertyToken(property_name, raw_text, tag_content_format, text_alignment, destructuring,
                             first), next_value

    @staticmethod
    def __try_split_tag_content(tag_content: str) -> tuple[bool, str, str, str]:
        property_name_and_destructuring = None
        tag_content_format = None
        tag_content_alignment = None

        def index_of(text: str, sub: str):
            try:
                return text.index(sub)
            except ValueError:
                return -1

        format_delim = index_of(tag_content, ':')
        alignment_delim = index_of(tag_content, ',')
        if format_delim == -1 and alignment_delim == -1:
            return True, tag_content, None, None
        if alignment_delim == -1 or (format_delim != -1 and alignment_delim > format_delim):
            property_name_and_destructuring = tag_content[:format_delim]
            tag_content_format = None if format_delim == len(tag_content) - 1 else tag_content[format_delim + 1:]
            tag_content_alignment = None
            return True, property_name_and_destructuring, tag_content_format, tag_content_alignment
        property_name_and_destructuring = tag_content[:alignment_delim]
        if format_delim == -1:
            if alignment_delim == len(tag_content) - 1:
                tag_content_alignment = None
                tag_content_format = None
                return False, property_name_and_destructuring, tag_content_format, tag_content_alignment
            tag_content_format = None
            tag_content_alignment = tag_content[alignment_delim + 1:]
            return True, property_name_and_destructuring, tag_content_format, tag_content_alignment
        if alignment_delim == format_delim - 1:
            return False, property_name_and_destructuring, tag_content_format, tag_content_alignment

        tag_content_alignment = tag_content[alignment_delim + 1: format_delim ]
        if format_delim == len(tag_content) - 1:
            tag_content_format = None
        else:
            tag_content_format = tag_content[format_delim + 1:]
        return True, property_name_and_destructuring, tag_content_format, tag_content_alignment

    @staticmethod
    def __is_valid_in_property_tag(ch: str) -> bool:
        return MessageTemplateParser.__is_valid_in_destructuring_hint(ch) or \
               MessageTemplateParser.__is_valid_in_property_name(ch) or \
               MessageTemplateParser.__is_valid_in_format(ch) or \
               ch == ':'

    @staticmethod
    def __is_valid_in_destructuring_hint(ch: str) -> bool:
        return ch == '@' or ch == '$'

    @staticmethod
    def __is_valid_in_property_name(ch: str) -> bool:
        return ch == '_' or ch.isalpha() or ch.isdigit()

    @staticmethod
    def __is_valid_in_format(ch: str) -> bool:
        return ch != '}' and ch != '$' \
               and (ch.isalpha() or ch.isdigit() or ch in string.punctuation or ch == ' ' or ch == '+')

    @staticmethod
    def __is_valid_in_alignment(ch: str):
        return ch.isdigit() or ch == '-'

    @staticmethod
    def __try_get_destructuring_hint(ch: str) -> tuple[bool, Destructuring]:
        match ch:
            case '@':
                return True, Destructuring.DESTRUCTURE
            case '$':
                return True, Destructuring.STRINGIFY
            case _:
                return False, Destructuring.DEFAULT



from typing import Callable

import pytest

from pyserilog.core.enrichers.conditional_enricher import ConditionalEnricher
from support.collecting_enricher import CollectingEnricher
from test_dummies import *
from dummy_wrapping_sink import DummyWrappingSink
from pyserilog.logger_configuration import LoggerConfiguration
from pyserilog.configuration.logger_sink_configuration import LoggerSinkConfiguration
from pyserilog.configuration.logger_enrichment_configuration import LoggerEnrichmentConfiguration
from pyserilog.core.constants import Constants
from pyserilog.core.ilog_event_property_value_factory import ILogEventPropertyValueFactory
from pyserilog.core.ilog_event_sink import ILogEventSink
from pyserilog.core.logger import Logger
from pyserilog.core.logging_level_switch import LoggingLevelSwitch
from pyserilog.core.filter.delegate_filter import DelegateFilter
from pyserilog.events.log_event import LogEvent
from pyserilog.events.log_event_level import LogEventLevel
from pyserilog.events.scalar_value import ScalarValue
from pyserilog.events.structure_value import StructureValue
from pyserilog.policies.ProjectedDestructuringPolicy import ProjectedDestructuringPolicy
from support.collecting_sink import CollectingSink
from support.delegating_enricher import DelegatingEnricher
from support.some import Some
from support.delegating_sink import DelegatingSink
from support.string_sink import StringSink


class AB:
    def __init__(self):
        self._a: int = 0
        self._b: int = 0

    @property
    def a(self):
        return self._a

    @a.setter
    def a(self, value):
        self._a = value

    @property
    def b(self):
        return self._b

    @b.setter
    def b(self, value):
        self._b = value


def log_and_get_as_string(x, config_func: Callable[[LoggerConfiguration], LoggerConfiguration],
                          destructuring_symbol: str = ""):
    evt = {'res': None}

    def write_func(log_event: LogEvent):
        evt['res'] = log_event

    log_conf = LoggerConfiguration() \
        .write_to.sink(DelegatingSink(write_func))
    log_conf = config_func(log_conf)
    log = log_conf.create_logger()

    log.information(f"{{{destructuring_symbol}X}}", x)

    assert evt['res'] is not None
    res = evt['res']
    assert isinstance(res, LogEvent)
    return str(res.properties["X"])


class ToStringOfLength:

    def __init__(self, to_string_of_length: int):
        self._to_string_of_length = to_string_of_length

    def __str__(self):
        return "#" * self._to_string_of_length


class Expando(object):
    pass


class Value:
    pass


class DisposableSink(ILogEventSink):
    def __init__(self):
        self.is_disposed = False

    def emit(self, log_event: LogEvent):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.is_disposed = True


class TestLoggerConfiguration:
    class DisposableSink(ILogEventSink):

        def __init__(self):
            self._is_disposed = False

        @property
        def is_disposed(self):
            return self._is_disposed

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            self._is_disposed = True

        def emit(self, log_event: LogEvent):
            pass

    def test_disposable_sinks_are_disposed_along_with_root_logger(self):
        sink = TestLoggerConfiguration.DisposableSink()
        with LoggerConfiguration() \
                .write_to \
                .sink(sink) \
                .create_logger():
            # Do nothing
            assert True

        assert sink.is_disposed

    def test_disposable_sinks_are_not_disposed_along_with_contextual_loggers(self):
        sink = DisposableSink()
        with LoggerConfiguration() \
                .write_to.sink(sink) \
                .create_logger() \
                .for_context(class_type=TestLoggerConfiguration):
            pass

        assert not sink.is_disposed

    def test_a_filter_prevents_matched_events_from_passing_to_the_sink(self):
        excluded = Some.information_event()
        included = Some.information_event()

        delegate_filter = DelegateFilter(lambda e: e.message_template != excluded.message_template)
        events = []
        sink = DelegatingSink(events.append)
        logger = LoggerConfiguration() \
            .write_to.sink(sink) \
            .filter.with_filter(delegate_filter) \
            .create_logger()
        logger.write_event(included)
        logger.write_event(excluded)

        assert len(events) == 1
        assert included in events

    def test_specifying_that_a_type_is_scalar_causes_it_to_be_logged_as_scalar_even_when_destructuring(self):
        events = []
        sink = DelegatingSink(events.append)

        with LoggerConfiguration() \
                .write_to.sink(sink) \
                .destructure.as_scalar(AB) \
                .create_logger() as logger:
            logger.information("{@AB}", AB())

            assert len(events) == 1
            ev: LogEvent = events[0]
            prop = ev.properties["AB"]
            assert isinstance(prop, ScalarValue)

    def test_destructuring_system_type_gives_scalar_by_default(self):
        events: list[LogEvent] = []
        sink = DelegatingSink(events.append)

        logger = LoggerConfiguration() \
            .write_to.sink(sink) \
            .create_logger()

        this_type = TestLoggerConfiguration
        logger.information("{@this_type}", this_type)

        ev: LogEvent = events[0]
        prop = ev.properties["this_type"]

        assert isinstance(prop, ScalarValue)
        assert this_type == prop.value

    def test_destructuring_is_possible_for_system_type_derived_properties(self):
        events: list[LogEvent] = []
        sink = DelegatingSink(events.append)

        def can_apply_func(input_typ: type):
            return issubclass(type, input_typ)

        def projection_func(input_typ: type):
            return input_typ.__name__

        policy = ProjectedDestructuringPolicy(can_apply_func=can_apply_func, projection_func=projection_func)
        logger = LoggerConfiguration() \
            .destructure \
            .with_policies(policy) \
            .write_to.sink(sink) \
            .create_logger()

        this_type = TestLoggerConfiguration
        logger.information("{@this_type}", this_type)

        ev = events[0]
        prop = ev.properties["this_type"]
        assert isinstance(prop, ScalarValue)
        assert this_type.__name__ == prop.value

    def test_transformations_are_applied_to_event_properties(self):
        events: list[LogEvent] = []
        sink = DelegatingSink(events.append)

        class TestC:
            def __init__(self, value):
                self._c_property = value

            @property
            def c_property(self):
                return self._c_property

            @c_property.setter
            def c_property(self, value):
                self._c_property = value

        def create_test(k):
            return TestC(k)

        logger = LoggerConfiguration() \
            .write_to.sink(sink) \
            .destructure \
            .by_transforming(AB, create_test) \
            .create_logger()

        logger.information("{@AB}", AB())

        ev = events[0]
        prop = ev.properties["AB"]

        assert isinstance(prop, StructureValue)
        sv = prop.properties[0]
        assert "c_property" == sv.name

    def test_writing_to_a_logger_writes_to_sub_logger(self):
        event_received = {'res': False}

        def configure_logger_function(l: LoggerConfiguration):
            def set_val(k):
                event_received['res'] = True

            l.write_to.sink(DelegatingSink(set_val))

        logger = LoggerConfiguration() \
            .write_to \
            .logger(configure_logger_function) \
            .create_logger()

        logger.write_event(Some.information_event())

        assert event_received['res']

    def test_sub_logger_restricts_filter(self):
        event_received = {'res': False}

        def configure_logger_function(l: LoggerConfiguration):
            def set_val(k):
                event_received['res'] = True

            log: LoggerConfiguration = l.minimum_level.fatal()
            log.write_to.sink(DelegatingSink(set_val))

        logger = LoggerConfiguration() \
            .write_to.logger(configure_logger_function) \
            .create_logger()

        logger.write_event(Some.information_event())

        assert not event_received['res']

    def test_enrichers_execute_in_configuration_order(self):
        prop = Some.log_event_property()
        enriched_property_seen = {'res': False}

        def first(log_event: LogEvent, factory: ILogEventPropertyValueFactory):
            t = factory
            log_event.add_property_if_absent(prop)

        def second(log_event: LogEvent, factory: ILogEventPropertyValueFactory):
            t = factory
            enriched_property_seen['res'] = prop.name in log_event.properties

        logger_conf: LoggerConfiguration = LoggerConfiguration().write_to.sink(StringSink())
        logger_conf: LoggerConfiguration = logger_conf.enrich.with_enrichers(DelegatingEnricher(first))
        logger: Logger = logger_conf.enrich.with_enrichers(DelegatingEnricher(second)).create_logger()

        logger.write_event(Some.information_event())

        assert enriched_property_seen['res']

    def test_maximum_destructuring_depth_default_is_effective(self):
        x = Expando()
        x.lvl1 = Expando()
        x.lvl1.lvl2 = Expando()
        x.lvl1.lvl2.lvl3 = Expando()
        x.lvl1.lvl2.lvl3.lvl4 = Expando()
        x.lvl1.lvl2.lvl3.lvl4.lvl5 = Expando()
        x.lvl1.lvl2.lvl3.lvl4.lvl5.lvl6 = Expando()
        x.lvl1.lvl2.lvl3.lvl4.lvl5.lvl6.lvl7 = Expando()
        x.lvl1.lvl2.lvl3.lvl4.lvl5.lvl6.lvl7.lvl8 = Expando()
        x.lvl1.lvl2.lvl3.lvl4.lvl5.lvl6.lvl7.lvl8.lvl9 = Expando()
        x.lvl1.lvl2.lvl3.lvl4.lvl5.lvl6.lvl7.lvl8.lvl9.lvl10 = Expando()
        x.lvl1.lvl2.lvl3.lvl4.lvl5.lvl6.lvl7.lvl8.lvl9.lvl10 = "lvl11"

        xs = log_and_get_as_string(x, lambda conf: conf, "@")

        assert "lvl10" in xs
        assert "lvl11" not in xs

    def test_maximum_destructuring_depth_is_effective(self):
        x = Expando()
        x.a_filed = Expando()
        x.a_filed.b_filed = Expando()
        x.a_filed.b_filed.c_filed = Expando()
        x.a_filed.b_filed.c_filed.d_field = Expando()
        x.a_filed.b_filed.c_filed.d_field = "f"

        def config_func(conf: LoggerConfiguration):
            return conf.destructure.to_maximum_depth(3)

        xs = log_and_get_as_string(x, config_func, "@")

        assert "c_filed" in xs
        assert "d_filed" not in xs

    def test_maximum_string_length_throws_for_limit_lower_than2(self):
        with pytest.raises(ValueError) as ex:
            LoggerConfiguration().destructure.to_maximum_string_length(1)

    def test_maximum_string_length_not_effective_for_string(self):
        x = "ABCD"
        limited_text = log_and_get_as_string(x, lambda conf: conf.destructure.to_maximum_string_length(3))

        assert "\"ABCD\"" == limited_text

    def test_maximum_string_length_effective_for_captured_string(self):
        x = "ABCD"

        limited_text = log_and_get_as_string(x, lambda conf: conf.destructure.to_maximum_string_length(3), "@")

        assert "\"AB…\"" == limited_text

    def test_maximum_string_length_effective_for_stringified_string(self):
        x = "ABCD"

        limited_text = log_and_get_as_string(x, lambda conf: conf.destructure.to_maximum_string_length(3), "$")

        assert "\"AB…\"" == limited_text

    @pytest.mark.parametrize("text,text_after,limit", [("1234", "12…", 3), ("123", "123", 3)])
    def test_maximum_string_length_effective_for_captured_object(self, text: str, text_after: str, limit: int):
        x = Expando()
        x.too_long_text = text

        limited_text = log_and_get_as_string(x, lambda conf: conf.destructure.to_maximum_string_length(limit), "@")

        assert text_after in limited_text

    def test_maximum_string_length_effective_for_stringified_object(self):
        x = ToStringOfLength(4)

        limited_text = log_and_get_as_string(x, lambda conf: conf.destructure.to_maximum_string_length(3), "$")
        assert "##…" in limited_text

    def test_maximum_string_length_not_effective_for_object(self):
        x = ToStringOfLength(4)

        limited_text = log_and_get_as_string(x, lambda conf: conf.destructure.to_maximum_string_length(3))

        assert "####" in limited_text

    def test_maximum_string_collection_throws_for_limit_lower_than1(self):
        with pytest.raises(ValueError) as ve:
            LoggerConfiguration().destructure.to_maximum_collection_count(0)

    def test_maximum_collection_count_not_effective_for_array_as_long_as_limit(self):
        x = [1, 2, 3]

        limited_collection = log_and_get_as_string(x, lambda conf: conf.destructure.to_maximum_collection_count(3))

        assert "3" in limited_collection

    def test_maximum_collection_count_effective_for_array_than_limit(self):
        x = [1, 2, 3, 4]

        limited_collection = log_and_get_as_string(x, lambda conf: conf.destructure.to_maximum_collection_count(3))

        assert "3" in limited_collection
        assert "4" not in limited_collection

    def test_maximum_collection_count_effective_for_dictionary_with_more_keys_than_limit(self):
        x = {
            "1": 1,
            "2": 2,
            "3": 3
        }

        limited_collection = log_and_get_as_string(x, lambda conf: conf.destructure.to_maximum_collection_count(2))

        assert "2" in limited_collection
        assert "3" not in limited_collection

    def test_maximum_collection_count_not_effective_for_dictionary_with_as_many_keys_as_limit(self):
        x = {
            "1": 1,
            "2": 2,
        }

        limited_collection = log_and_get_as_string(x, lambda conf: conf.destructure.to_maximum_collection_count(2))

        assert "2" in limited_collection

    def test_dynamically_switching_sink_restricts_output(self):
        events_received = {'res': 0}
        level_switch = LoggingLevelSwitch()

        def write_func(log_event: LogEvent):
            events_received['res'] += 1

        logger = LoggerConfiguration() \
            .write_to.sink(DelegatingSink(write_func), level_switch=level_switch) \
            .create_logger()

        logger.write_event(Some.information_event())
        level_switch.minimum_level = LogEventLevel.WARNING
        logger.write_event(Some.information_event())
        level_switch.minimum_level = LogEventLevel.VERBOSE
        logger.write_event(Some.information_event())

        assert 2 == events_received['res']

    def test_level_switch_takes_precedence_over_minimum_level(self):
        sink = CollectingSink()

        logger = LoggerConfiguration() \
            .write_to.sink(sink, LogEventLevel.FATAL, LoggingLevelSwitch()) \
            .create_logger()

        logger.write_event(Some.information_event())

        assert len(sink.events) == 1

    def test_last_minimum_level_configuration_wins(self):
        sink = CollectingSink()

        logger_conf: LoggerConfiguration = LoggerConfiguration() \
            .minimum_level.controlled_by(LoggingLevelSwitch(LogEventLevel.FATAL))
        logger_conf: LoggerConfiguration = logger_conf.minimum_level.debug()
        logger: Logger = logger_conf.write_to.sink(sink) \
            .create_logger()

        logger.write_event(Some.information_event())

        assert len(sink.events) == 1

    def test_higher_minimum_level_overrides_are_propagated(self):
        sink = CollectingSink()

        logger_conf: LoggerConfiguration = LoggerConfiguration() \
            .minimum_level.debug()
        logger_conf: LoggerConfiguration = logger_conf.minimum_level.override("Microsoft", LogEventLevel.FATAL)
        logger = logger_conf.write_to.sink(sink).create_logger()

        logger.write_event(Some.information_event())
        logger.for_context(property_name=Constants.SOURCE_CONTEXT_PROPERTY_NAME, value="Microsoft.AspNet.Something") \
            .write_event(Some.information_event())
        logger.for_context(class_type=TestLoggerConfiguration).write_event(Some.information_event())

        assert 2 == len(sink.events)

    def test_lower_minimum_level_overrides_are_propagated(self):
        sink = CollectingSink()

        logger: Logger = LoggerConfiguration() \
            .minimum_level.error() \
            .minimum_level.override("Microsoft", LogEventLevel.DEBUG) \
            .write_to.sink(sink) \
            .create_logger()

        logger.write_event(Some.information_event())
        logger.for_context(property_name=Constants.SOURCE_CONTEXT_PROPERTY_NAME, value="Microsoft.AspNet.Something") \
            .write_event(Some.information_event())
        logger.for_context(class_type=TestLoggerConfiguration) \
            .write_event(Some.information_event())

        assert 1 == len(sink.events)

    def test_exceptions_thrown_by_sinks_are_not_propagated(self):
        def func(x):
            raise Exception("Boom!")

        logger = LoggerConfiguration() \
            .write_to.sink(DelegatingSink(func)) \
            .create_logger()

        logger.write_event(Some.information_event())

        assert True, "No exception reached the caller"

    def test_exceptions_thrown_by_sinks_are_not_propagated_even_when_auditing_is_present(self):
        def func(x):
            raise Exception("Boom!")

        logger = LoggerConfiguration() \
            .audit_to.sink(CollectingSink()) \
            .write_to.sink(DelegatingSink(func)) \
            .create_logger()

        logger.write_event(Some.information_event())

        assert True, "No exception reached the caller"

    def test_exceptions_thrown_by_filters_are_not_propagated(self):
        def func(x):
            raise Exception("Boom!")

        logger = LoggerConfiguration() \
            .filter.by_excluding(func) \
            .create_logger()

        logger.write_event(Some.information_event())

        assert True, "No exception reached the caller"

    def test_exceptions_thrown_by_destructuring_policies_are_not_propagated(self):
        def func(x):
            raise Exception("Boom!")

        logger = LoggerConfiguration() \
            .write_to.sink(CollectingSink()) \
            .destructure.by_transforming(Value, func) \
            .create_logger()

        logger.information("{@Value}", Value())

        assert True, "No exception reached the caller"

    def test_wrapping_decorates_the_configured_sink(self):
        DummyWrappingSink.reset()
        sink = CollectingSink()
        logger_conf = DummyLoggerConfigurationExtensions.dummy(LoggerConfiguration().write_to, lambda w: w.sink(sink))
        logger = logger_conf.create_logger()

        evt = Some.information_event()
        logger.write_event(evt)

        em: list = DummyWrappingSink.emitted()
        assert evt == em[0]
        assert evt == sink.single_event

    def test_wrapping_does_not_permit_enrichment(self):
        sink = CollectingSink()
        property_name = Some.string()
        logger = LoggerConfiguration().write_to.dummy(lambda w: w.sink(sink) \
                                                      .enrich.with_property(property_name, 1)) \
            .create_logger()

        evt = Some.information_event()
        logger.write_event(evt)

        assert evt == sink.single_event
        assert property_name not in evt.properties

    def test_wrapping_is_applied_when_chaining(self):
        DummyWrappingSink.reset()
        sink1 = CollectingSink()
        sink2 = CollectingSink()
        logger = LoggerConfiguration().write_to.dummy(lambda w: w.sink(sink1).write_to.sink(sink2)) \
            .create_logger()

        evt = Some.information_event()
        logger.write_event(evt)

        assert evt == DummyWrappingSink.emitted()[0]
        assert evt == sink1.single_event
        assert evt == sink2.single_event

    def test_wrapping_is_applied_when_calling_multiple_times(self):
        DummyWrappingSink.reset()
        sink1 = CollectingSink()
        sink2 = CollectingSink()

        def func(w: LoggerSinkConfiguration):
            w.sink(sink1)
            w.sink(sink2)

        logger = LoggerConfiguration().write_to.dummy(func) \
            .create_logger()

        evt = Some.information_event()
        logger.write_event(evt)

        assert evt == DummyWrappingSink.emitted()[0]
        assert evt == sink1.single_event
        assert evt == sink2.single_event

    def test_wrapping_sink_respects_log_event_level_setting(self):
        DummyWrappingSink.reset()
        sink = CollectingSink()
        logger = LoggerConfiguration() \
            .write_to.dummy_wrap(lambda w: w.sink(sink), LogEventLevel.ERROR, None) \
            .create_logger()

        logger.write_event(Some.information_event())

        assert len(DummyWrappingSink.emitted()) == 0
        assert 0 == len(sink.events)

    def test_wrapping_sink_respects_level_switch_setting(self):
        DummyWrappingSink.reset()
        sink = CollectingSink()
        logger = LoggerConfiguration() \
            .write_to.dummy_wrap(
            lambda w: w.sink(sink), LogEventLevel.VERBOSE,
            LoggingLevelSwitch(LogEventLevel.ERROR)) \
            .create_logger()

        logger.write_event(Some.information_event())

        assert len(DummyWrappingSink.emitted()) == 0
        assert len(sink.events) == 0

    def test_wrapping_sink_receives_events_when_level_is_appropriate(self):
        DummyWrappingSink.reset()
        sink = CollectingSink()
        logger = LoggerConfiguration() \
            .write_to.dummy_wrap(
            lambda w: w.sink(sink), LogEventLevel.ERROR,
            LoggingLevelSwitch(LogEventLevel.VERBOSE)) \
            .create_logger()

        logger.write_event(Some.information_event())

        assert len(DummyWrappingSink.emitted()) > 0
        assert len(sink.events) > 0

    def test_conditional_sinks_receive_events_matching_condition(self):
        matching = CollectingSink()
        logger = LoggerConfiguration() \
            .write_to.conditional(
            lambda le: le.level == LogEventLevel.WARNING,
            lambda w: w.sink(matching)) \
            .create_logger()

        logger.information("Information")
        logger.warning("Warning")
        logger.error("Error")

        evt = matching.single_event
        assert LogEventLevel.WARNING == evt.level

    def test_enrichers_can_be_wrapped(self):
        enricher = CollectingEnricher()

        configuration = LoggerConfiguration()
        LoggerEnrichmentConfiguration.wrap(
            configuration.enrich,
            lambda e: ConditionalEnricher(e, lambda le: le.level == LogEventLevel.WARNING),
            lambda enrich: enrich.with_enrichers(enricher))

        logger = configuration.create_logger()
        logger.information("Information")
        logger.warning("Warning")
        logger.error("Error")

        assert len(enricher.events) == 1
        evt = enricher.events[0]
        assert LogEventLevel.WARNING == evt.level

    def testConditionalEnrichersCheckConditions(self):
        enricher = CollectingEnricher()

        logger = LoggerConfiguration() \
            .enrich.when(lambda le: le.level == LogEventLevel.WARNING, lambda enrich: enrich.with_enrichers(enricher)) \
            .create_logger()

        logger.information("Information")
        logger.warning("Warning")
        logger.error("Error")

        assert 1 == len(enricher.events)
        evt = enricher.events[0]
        assert LogEventLevel.WARNING == evt.level

    def test_leveled_enrichers_check_levels(self):
        enricher = CollectingEnricher()

        logger = LoggerConfiguration() \
            .enrich.at_level(LogEventLevel.WARNING, lambda enrich: enrich.with_enrichers(enricher)) \
            .create_logger()

        logger.information("Information")
        logger.warning("Warning")
        logger.error("Error")

        assert 2 == len(enricher.events)
        assert 0 == len(list(filter(lambda x: not x.level >= LogEventLevel.WARNING, enricher.events)))

    def test_leveled_enrichers_check_level_switch(self):
        enricher = CollectingEnricher()
        level_switch = LoggingLevelSwitch(LogEventLevel.WARNING)

        logger = LoggerConfiguration() \
            .enrich.at_level(level_switch, lambda enrich: enrich.with_enrichers(enricher)) \
            .create_logger()

        logger.information("Information")
        logger.warning("Warning")
        logger.error("Error")

        assert 2 == len(enricher.events)
        assert 0 == len(list(filter(lambda x: not x.level >= LogEventLevel.WARNING, enricher.events)))

        enricher.events.clear()
        level_switch.minimum_level = LogEventLevel.ERROR

        logger.information("Information")
        logger.warning("Warning")
        logger.error("Error")

        assert 1 == len(enricher.events)
        error = enricher.events[0]
        assert error.level == LogEventLevel.ERROR



from pyserilog.core.idestructuring_policy import IDestructuringPolicy
from pyserilog.core.ilog_event_property_value_factory import ILogEventPropertyValueFactory
from pyserilog.events.log_event_property_value import LogEventPropertyValue
from pyserilog.events.scalar_value import ScalarValue


class ReflectionTypesScalarDestructuringPolicy(IDestructuringPolicy):
    def try_destructure(self, value, property_value_factory: ILogEventPropertyValueFactory) \
            -> tuple[bool, LogEventPropertyValue | None]:
        if not isinstance(value, type):
            return False, None
        return True, ScalarValue(value)



from typing import Callable

from pyserilog.core.ilog_event_sink import ILogEventSink
from pyserilog.events.log_event import LogEvent


class ConditionalSink(ILogEventSink):
    def __init__(self, wrapper: ILogEventSink, condition: Callable[[LogEvent], bool]):
        self._wrapper = wrapper
        self._condition = condition

    def emit(self, log_event: LogEvent):
        if self._condition(log_event):
            self._wrapper.emit(log_event)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if hasattr(self._wrapper, "__exit__"):
            self._wrapper.__exit__(exc_type, exc_val, exc_tb)



from pyserilog import LoggerConfiguration, Guard
from pyserilog.configuration.ilogger_settings import ILoggerSettings
from pyserilog.settings.key_value_paies.key_value_pair_settings import KeyValuePairSettings


class LoggerSettingsConfiguration:

    def __init__(self, logger_configuration):
        self._logger_configuration = logger_configuration

    def settings(self, settings: ILoggerSettings) -> LoggerConfiguration:
        Guard.against_null(settings)

        settings.configure(self._logger_configuration)
        return self._logger_configuration

    def key_value_pairs(self, settings: list[tuple[str, str]] | dict) -> LoggerConfiguration:
        Guard.against_null(settings)

        unique_settings = {}
        if isinstance(settings, list):
            for (key, value) in settings:
                unique_settings[key] = value
        elif isinstance(settings, dict):
            for key in settings:
                unique_settings[key] = settings[key]

        return self.__key_value_pairs(unique_settings)

    def __key_value_pairs(self, settings: dict):
        return self.settings(KeyValuePairSettings(settings))



from pyserilog.core.string_writable import StringWriteable
from typing import Callable

from pyserilog.guard import Guard
from pyserilog.events.log_event_property import LogEventProperty
from pyserilog.events.log_event_property_value import LogEventPropertyValue


class StructureValue(LogEventPropertyValue):

    def __init__(self, properties: list[LogEventProperty], type_tag: str | None = None):
        Guard.against_null(properties)

        self._properties = properties
        self._type_tag = type_tag

    @property
    def properties(self) -> list[LogEventProperty]:
        return self._properties

    @property
    def type_tag(self) -> str | None:
        return self._type_tag

    def render(self, output: StringWriteable, text_format: str | None = None, formatter: Callable[[object], str] = None):
        if self.type_tag is not None:
            output.write(str(self.type_tag))
            output.write(" ")
        output.write("{ ")
        for prop in self._properties[:-1]:
            self._render(prop, output)
            output.write(", ")
        if len(self.properties) > 0:
            last = self.properties[-1]

            self._render(last, output)
        output.write(" }")

    @staticmethod
    def _render(prop: LogEventProperty, output: StringWriteable, text_format: str | None = None,
                formatter: Callable[[object], str] = None):
        output.write(prop.name)
        output.write(": ")
        prop.value.render(output, text_format, formatter)



from pyserilog.core.iscalar_conversion_policy import IScalarConversionPolicy
from pyserilog.events.scalar_value import ScalarValue


class SimpleScalarConversionPolicy(IScalarConversionPolicy):
    def __init__(self, scalar_types: list):
        self._scalar_type = set(scalar_types)

    def try_convert_to_scalar(self, value) -> tuple[bool, ScalarValue | None]:
        if type(value) in self._scalar_type:
            return True, ScalarValue(value)
        return False, None


from pyserilog.core.string_writable import StringIOWriteable,StringWriteable

from pyserilog.debuging.self_log import SelfLog
from pyserilog.events.log_event_property_value import LogEventPropertyValue
from pyserilog.parsing.message_template_token import MessageTemplateToken
from pyserilog.parsing.property_token import PropertyToken
from pyserilog.rendering.message_template_renderer import MessageTemplateRenderer


class MessageTemplate:
    def __init__(self, tokens: list[MessageTemplateToken], text: str | None = None):
        if text is None:
            text = ''.join(tokens)
        self._text = text
        self._tokens = tokens
        property_tokens: list[PropertyToken] = self.get_elements_of_type(tokens, PropertyToken)
        self._named_properties: list[PropertyToken] | None = None
        self._positional_properties = None
        if len(property_tokens) != 0:
            all_positional = True
            any_positional = False
            for property_token in property_tokens:
                if property_token.is_positional:
                    any_positional = True
                else:
                    all_positional = False
            if all_positional:
                self._positional_properties = property_tokens
            else:
                if any_positional:
                    SelfLog.write_line("Message template is malformed: {0}", text)
                self._named_properties = property_tokens

    @staticmethod
    def get_elements_of_type(tokens: list[MessageTemplateToken], element_type):
        result: list[PropertyToken] = []
        for token in tokens:
            if isinstance(token, element_type):
                result.append(token)
        return result

    @property
    def tokens(self):
        return self._tokens

    @property
    def text(self):
        return self._text

    @property
    def named_properties(self) -> list[PropertyToken] | None:
        return self._named_properties

    @property
    def positional_properties(self) -> list[PropertyToken] | None:
        return self._positional_properties

    def render(self, properties: dict[str, LogEventPropertyValue], output: StringWriteable | None = None) -> str | None:
        if output is None:
            return self.__render_without_writer(properties=properties)
        MessageTemplateRenderer.render(self, properties, output)

    def __render_without_writer(self, properties: dict[str, LogEventPropertyValue]) -> str:
        writer = StringIOWriteable()
        self.render(properties, writer)
        result = writer.getvalue()
        writer.close()
        return result

    def __str__(self):
        return self._text



from pyserilog.core.iscalar_conversion_policy import IScalarConversionPolicy
from pyserilog.events.scalar_value import ScalarValue


class ByteArrayScalarConversionPolicy(IScalarConversionPolicy):
    maximum_byte_array_length: int = 1024

    def try_convert_to_scalar(self, value) -> tuple[bool, ScalarValue | None]:
        if isinstance(value, bytearray) or \
                (isinstance(value, list) and len(value) > 0 and isinstance(value[0], bytes)):
            value = b''.join(value)

        if not isinstance(value, bytes):
            return False, None

        if len(value) > self.maximum_byte_array_length:
            start : str = value[0:16].hex().upper()
            hex_str = f"{start}… ({len(value)} bytes)"
        else:
            hex_str = value.hex()
        return True, ScalarValue(hex_str)



from pyserilog.core.ilog_event_property_factory import ILogEventPropertyFactory
from pyserilog.core.pipeline.message_template_cache import MessageTemplateCache
from pyserilog.events.message_template import MessageTemplate
from pyserilog.events.event_property import EventProperty
from pyserilog.parsing.message_template_parser import MessageTemplateParser
from pyserilog.capturing.property_value_converter import PropertyValueConverter
from pyserilog.capturing.property_binder import PropertyBinder


class MessageTemplateProcessor(ILogEventPropertyFactory):
    __parser = MessageTemplateCache(MessageTemplateParser())

    def __init__(self, property_value_converter: PropertyValueConverter):
        self._property_value_converter = property_value_converter
        self._property_binder = PropertyBinder(self._property_value_converter)

    def create_property(self, name: str, value, destructure_objects: bool = False):
        return self._property_value_converter.create_property(name, value, destructure_objects)

    def create_property_value(self, value, destructure_objects: bool = False ):
        return self._property_value_converter.create_property_value_check_destructuring(value , destructure_objects)

    def process(self, message_template: str, message_template_parameters: list) -> \
            tuple[MessageTemplate, list[EventProperty]]:
        parsed_template = self.__parser.parse(message_template)
        properties = self._property_binder.construct_properties(parsed_template, message_template_parameters)
        return parsed_template, properties
